#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Coverage.py


from concurrent.futures import ThreadPoolExecutor
from math import ceil
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
from . import RPFs


class Coverage(object):

    def __init__(self, args):
        # input and output
        self.rpf = args.rpf
        self.gene = args.transcript
        self.gene_table = None
        self.output = args.output

        # filter the stable codon coverage range
        self.tis = None
        self.tts = None
        self.frame = args.frame
        self.site = 'all'
        self.rpf_num = args.min
        bins = args.bin.split(',')
        self.utr5_bin = int(bins[0])
        self.cds_bin = int(bins[1])
        self.utr3_bin = int(bins[2])

        self.norm = args.normal
        self.outlier = args.outlier
        self.thread = args.thread
        self.set = args.set
        self.sample_name = None

        # for the high expression gene filter
        self.raw_rpf = None
        self.merged_rpf = None
        # self.high_gene = None
        self.high_rpf = None
        self.gene_num = None
        self.total_rpf_num = None

        # mean of coverage
        self.utr5_dict = OrderedDict()
        self.cds_dict = OrderedDict()
        self.utr3_dict = OrderedDict()
        self.utr5_mean = None
        self.cds_mean = None
        self.utr3_mean = None

    def read_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis=self.tis,
                                      tts=self.tts,
                                      gene=self.gene,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1].copy()
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        self.gene_num = len(self.high_gene)
        self.high_rpf = rpf_results[7].copy()
        del rpf_results

        if self.norm:
            self.high_rpf.loc[:, self.sample_name] = self.high_rpf[self.sample_name] * 1e6 / self.total_rpf_num


    def import_gene(self):
        '''
        @Message  : import the gene file
        @Input    : self.gene --> gene file
        @Return   : self.gene --> gene dataframe
        @Flow     : step1 --> import the gene file
        '''

        self.gene_table = pd.read_csv(self.gene, sep='\t', index_col=False)
        self.gene_table = self.gene_table.loc[self.gene_table['transcript_id'].isin(self.high_gene), ]


    @staticmethod
    def run_fill_outliers(args):

        def iqr_fill_outliers(groups):
            q1 = groups.quantile(0.25)
            q3 = groups.quantile(0.75)
            iqr = q3 - q1
            lower_bound = q1 - 2 * iqr
            upper_bound = q3 + 2 * iqr
            groups[(groups < lower_bound) | (groups > upper_bound)] = groups.mean()
            return groups

        def zscore_fill_outliers(groups):
            z_scores = (groups - groups.mean()) / groups.std()
            threshold = 5
            groups[abs(z_scores) > threshold] = groups.mean()
            # groups[abs(z_scores) > threshold] = groups.mean()
            return groups
        
        gene_df = args
        return gene_df.groupby('Gene').apply(lambda x: x.apply(iqr_fill_outliers))
    
    
    @staticmethod
    def filter_outliers(self, gene_coverage):
        '''
        @Message  : filter the outliers
        @Input    : self.utr5_mean, self.cds_mean, self.utr3_mean --> the mean of the gene rpfs dataframe
        @Return   : self.utr5_mean, self.cds_mean, self.utr3_mean --> the mean of the gene rpfs dataframe
        @Flow     : step1 --> calculate the z-score
                    step2 --> filter the outliers
        '''

        gene_cover_df_n = gene_coverage.drop(columns=['Bins'])
        
        # stat the gene number
        gene_list = list(self.high_gene)
        gene_list_len = len(self.high_gene)

        gene_list_split = [gene_list[i:i + gene_list_len // self.thread] for i in range(0, gene_list_len, ceil(gene_list_len / self.thread))]
        gene_cover_split = [gene_cover_df_n[gene_cover_df_n['Gene'].isin(gene_list_split[i])] for i in range(self.thread)]

        args = [(gene_cover_split[i]) for i in range(self.thread)]

        # run the fill_outliers function with multi-thread
        from multiprocessing import Pool
        pool = Pool(processes=self.thread)
        gene_cover_list = pool.map(self.run_fill_outliers, args)
        pool.close()
        pool.join()

        # gene_cover_df_n = gene_cover_df_n.groupby('Gene').apply(lambda x: x.apply(fill_outliers))
        gene_cover_df_n = pd.concat(gene_cover_list, axis=0)
        gene_cover_df_n.reset_index(inplace=True)
        gene_cover_df_n.rename(columns={'level_1': 'Bins'}, inplace=True)
        gene_cover_df_n['Bins'] = gene_coverage['Bins']
        gene_coverage = gene_cover_df_n.groupby('Bins')[self.sample_name].sum() / self.gene_num
        
        return gene_coverage


    def adjust_dimension(self, rpm, bins):
        '''
        @Message  : adjust the dimension of the gene rpfs dataframe
        @Input    : rpm --> gene rpfs dataframe contain the high expression genes
        @Return   : self.utr5_dict, self.cds_dict, self.utr3_dict --> the mean of the gene rpfs dataframe
                    self.utr5_mean, self.cds_mean, self.utr3_mean --> the mean of the gene rpfs dataframe
        @Flow     : step1 --> set the bins
                    step2 --> calculate the mean of the each gene rpfs
                    step3 --> calculate the mean of the gene rpfs in 5-UTR, CDS, 3-UTR
        '''
        
        region_dict = OrderedDict()
        
        for idx, gene in rpm.groupby('name'):
            gene.loc[:, 'Bins'] = pd.cut(gene.index, bins, labels=range(bins))
            # region_dict[idx] = gene.groupby('Bins')[self.sample_name].mean()
            # FutureWarning: The default of observed=False is deprecated and will be changed to True in a future version of pandas. 
            # Pass observed=False to retain current behavior or observed=True to adopt the future default and silence this warning.
            region_dict[idx] = gene.groupby('Bins', observed=True)[self.sample_name].mean()
        
        region_df = pd.concat(region_dict, axis=0)
        region_df.reset_index(inplace=True)
        region_df.rename(columns={'level_0': 'Gene'}, inplace=True)

        return region_df
    

    def process_utr5(self):

        print('Processing 5-UTR.', flush=True)

        # filter utr5 length
        fit_gene = self.gene_table.loc[self.gene_table['utr5_length'] > self.utr5_bin * 2, ]
        fit_gene_num = fit_gene.shape[0]

        if fit_gene_num == 0:
            # create the 0 dictionary with the same dimension
            self.utr5_dict = OrderedDict()
            for idx, gene in self.high_rpf.groupby('name'):
                self.utr5_dict[idx] = pd.DataFrame(np.zeros((self.utr5_bin, self.sample_num)), index=range(self.utr5_bin), columns=self.sample_name)

            # create the 0 dataframe for the output
            self.utr5_df = pd.concat(self.utr5_dict, axis=0).reset_index()
            self.utr5_df.rename(columns={'level_0': 'Gene', 'level_1': 'Bins'}, inplace=True)
            self.utr5_mean = pd.DataFrame(np.zeros((self.utr5_bin, self.sample_num)), index=range(self.utr5_bin), columns=self.sample_name)
            self.utr5_mean.index.name = 'Bins'

            return
        else:
            print('The number of gene with 5-UTR %d.' % fit_gene_num, flush=True)

        # calculate the mean of the gene rpfs in 5-UTR
        utr5_rpm = self.high_rpf.loc[self.high_rpf['region'] == '5utr', ]
        utr5_rpm = utr5_rpm.loc[utr5_rpm['name'].isin(fit_gene['transcript_id']), ]
        self.utr5_df = self.adjust_dimension(rpm = utr5_rpm, bins = self.utr5_bin)

        # check the outlier
        if self.outlier:
            print('Filter the outliers.', flush=True)
            self.utr5_mean = self.filter_outliers(self, self.utr5_df)
        else:
            print('No outliers filter.', flush=True)
            self.utr5_mean = self.utr5_df.groupby('Bins')[self.sample_name].sum() / self.gene_num


    def process_cds(self):

        print('Processing CDS.', flush=True)

        # filter cds length
        fit_gene = self.gene_table.loc[self.gene_table['cds_length'] > self.cds_bin * 2, ]
        fit_gene_num = fit_gene.shape[0]

        if fit_gene_num == 0:
            # create the 0 dictionary with the same dimension
            self.cds_dict = OrderedDict()
            for idx, gene in self.high_rpf.groupby('name'):
                self.cds_dict[idx] = pd.DataFrame(np.zeros((self.cds_bin, self.sample_num)), index=range(self.cds_bin), columns=self.sample_name)

            # create the 0 dataframe for the output
            self.cds_df = pd.concat(self.cds_dict, axis=0).reset_index()
            self.cds_df.rename(columns={'level_0': 'Gene', 'level_1': 'Bins'}, inplace=True)
            self.cds_mean = pd.DataFrame(np.zeros((self.cds_bin, self.sample_num)), index=range(self.cds_bin), columns=self.sample_name)
            self.cds_mean.index.name = 'Bins'

            return
        else:
            print('The number of gene with CDS %d.' % fit_gene_num, flush=True)

        # calculate the mean of the gene rpfs in CDS
        cds_rpm = self.high_rpf.loc[self.high_rpf['region'] == 'cds', ]
        cds_rpm = cds_rpm.loc[cds_rpm['name'].isin(fit_gene['transcript_id']), ]
        self.cds_df = self.adjust_dimension(rpm = cds_rpm, bins = self.cds_bin)
        self.cds_df.rename(columns={'level_0': 'Gene', 'level_1': 'Bins'}, inplace=True)

        # check the outlier
        if self.outlier:
            print('Filter the outliers.', flush=True)
            self.cds_mean = self.filter_outliers(self, self.cds_df)
        else:
            print('No outliers filter.', flush=True)
            self.cds_mean = self.cds_df.groupby('Bins')[self.sample_name].sum() / self.gene_num
        

    def process_utr3(self):

        print('Processing 3-UTR.', flush=True)

        # filter utr3 length
        fit_gene = self.gene_table.loc[self.gene_table['utr3_length'] > self.utr3_bin * 2, ]
        fit_gene_num = fit_gene.shape[0]

        if fit_gene_num == 0:
            # create the 0 dictionary with the same dimension
            self.utr3_dict = OrderedDict()
            for idx, gene in self.high_rpf.groupby('name'):
                self.utr3_dict[idx] = pd.DataFrame(np.zeros((self.utr3_bin, self.sample_num)), index=range(self.utr3_bin), columns=self.sample_name)

            # create the 0 dataframe for the output
            self.utr3_df = pd.concat(self.utr3_dict, axis=0).reset_index()
            self.utr3_df.rename(columns={'level_0': 'Gene', 'level_1': 'Bins'}, inplace=True)
            self.utr3_mean = pd.DataFrame(np.zeros((self.utr3_bin, self.sample_num)), index=range(self.utr3_bin), columns=self.sample_name)
            self.utr3_mean.index.name = 'Bins'

            return
        else:
            print('The number of gene with 3-UTR %d.' % fit_gene_num, flush=True)

        # calculate the mean of the gene rpfs in 3-UTR
        utr3_rpm = self.high_rpf.loc[self.high_rpf['region'] == '3utr', ]
        utr3_rpm = utr3_rpm.loc[utr3_rpm['name'].isin(fit_gene['transcript_id']), ]
        self.utr3_df = self.adjust_dimension(rpm = utr3_rpm, bins = self.utr3_bin)

        # check the outlier
        if self.outlier:
            print('Filter the outliers.', flush=True)
            self.utr3_mean = self.filter_outliers(self, self.utr3_df)
        else:
            print('No outliers filter.', flush=True)
            self.utr3_mean = self.utr3_df.groupby('Bins')[self.sample_name].sum() / self.gene_num


    def output_meta_gene(self):
        # output the mean coverage of
        self.utr5_mean.insert(0, 'Region', '5-UTR')
        self.cds_mean.insert(0, 'Region', 'CDS')
        self.utr3_mean.insert(0, 'Region', '3-UTR')

        self.utr5_mean.reset_index(inplace=True)
        self.cds_mean.reset_index(inplace=True)
        self.utr3_mean.reset_index(inplace=True)

        self.utr5_mean['Bins'] = list(range(self.utr5_bin))
        self.cds_mean['Bins'] = list(range(self.utr5_bin, self.utr5_bin + self.cds_bin))
        self.utr3_mean['Bins'] = list(range(self.utr5_bin + self.cds_bin, self.utr5_bin + self.cds_bin + self.utr3_bin))

        utr5_mean_melt = pd.melt(self.utr5_mean, id_vars=['Bins', 'Region'], var_name='Sample', value_name='Density')
        cds_mean_melt = pd.melt(self.cds_mean, id_vars=['Bins', 'Region'], var_name='Sample', value_name='Density')
        utr3_mean_melt = pd.melt(self.utr3_mean, id_vars=['Bins', 'Region'], var_name='Sample', value_name='Density')

        mean_coverage = pd.concat([utr5_mean_melt, cds_mean_melt, utr3_mean_melt], ignore_index=True)
        mean_coverage.sort_values(by=['Sample', 'Bins'], inplace=True)
        mean_coverage = mean_coverage[['Sample', 'Region', 'Bins', 'Density']]
        mean_coverage.to_csv(self.output + '_utr5_cds_utr3_mean_coverage.txt', sep='\t', index=False)
        
    def draw_meta_gene_line(self):
        merge_coverage = pd.concat([self.utr5_mean, self.cds_mean, self.utr3_mean], ignore_index=True)
        merge_coverage.index = merge_coverage.index + 1

        for sp in self.sample_name:
            out_pdf = sp + "_coverage_line_plot.pdf"
            out_png = sp + "_coverage_line_plot.png"
            print('Draw {sp} line plot.'.format(sp=sp), flush=True)
            matplotlib.use('AGG')
            fig = plt.figure(figsize=(8, 5), dpi=300)
            plt.plot(merge_coverage[sp], label=sp)
            plt.legend()
            plt.xticks([1, self.utr5_bin, self.utr5_bin + self.cds_bin, self.utr5_bin + self.cds_bin + self.utr3_bin], ['TSS', 'TIS', 'TTS', 'TES'])
            plt.title("Mean coverage of ({number} genes)".format(number=self.gene_num), fontsize=16)
            if self.norm:
                plt.ylabel('Mean coverage (RPM)', fontsize=14)
            else:
                plt.ylabel('Mean coverage (RPFs)', fontsize=14)
            plt.xlabel('position of genes', fontsize=14)
            fig.tight_layout()
            # plt.show()

            fig.savefig(fname=out_pdf)
            fig.savefig(fname=out_png)
            plt.close()

    def draw_meta_gene_heat(self):
        
        utr5_df = self.utr5_df.reset_index().rename(columns={'level_0': 'Gene'})
        cds_df = self.cds_df.reset_index().rename(columns={'level_0': 'Gene'})
        utr3_df = self.utr3_df.reset_index().rename(columns={'level_0': 'Gene'})

        for sp in self.sample_name:
            print('Draw {sp} heatmap.'.format(sp=sp), flush=True)
            utr5_sp = pd.pivot_table(utr5_df.loc[:, ['Gene', "Bins", sp]],
                                     index='Gene',
                                     values=sp,
                                     columns='Bins')
            cds_sp = pd.pivot_table(cds_df.loc[:, ['Gene', "Bins", sp]],
                                     index='Gene',
                                     values=sp,
                                     columns='Bins')
            utr3_sp = pd.pivot_table(utr3_df.loc[:, ['Gene', "Bins", sp]],
                                     index='Gene',
                                     values=sp,
                                     columns='Bins')
            coverage_sp = pd.concat([utr5_sp, cds_sp, utr3_sp], axis=1, ignore_index=True).fillna(0)
            coverage_sp.index.name = 'Gene'
            coverage_sp = coverage_sp.reindex(coverage_sp.mean(axis=1).sort_values(ascending=False).index, axis=0)

            gene_bins = '{utr5}_{cds}_{utr3}'.format(utr5=str(self.utr5_bin),
                                                     cds=str(self.cds_bin),
                                                     utr3=str(self.utr3_bin))
            file_name = "{sp}_{bins}_coverage.txt".format(sp=sp, bins=gene_bins)
            coverage_sp.to_csv(file_name, sep='\t', index=True)

            # out_pdf = sp + "_heat_plot.pdf"
            out_png = sp + "_" + gene_bins + "_heat_plot.png"

            matplotlib.use('AGG')
            fig = plt.figure(figsize=(8, 4 + self.gene_num / 500), dpi=300)
            # sns.clustermap(np.log2(coverage_sp + 1), col_cluster=False, cmap="YlGnBu")
            sns.heatmap(np.log2(coverage_sp + 1), cmap="YlGnBu")
            plt.xticks([1, self.utr5_bin, self.utr5_bin + self.cds_bin, self.utr5_bin + self.cds_bin + self.utr3_bin], ['TSS', 'TIS', 'TTS', 'TES'])
            plt.title("Mean coverage of ({number} genes)".format(number=self.gene_num))
            fig.tight_layout()
            # plt.show()

            # fig.savefig(fname=out_pdf)
            fig.savefig(fname=out_png)
            plt.close()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_odd_ratio.py


import pandas as pd
import argparse


def merge_odd_ratio_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the codon odd ratio files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for codon odd ratio files (e.g., '*_sum_codon_odd_ratio.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name (default: prefix + _codon_odd_ratio.txt).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_odd_ratio_files(odd_ratio_list):
    '''
    @Message  : retrieve the codon odd ratio table.
    @Input    : list --> pattern for codon odd ratio files (e.g., '*_sum_codon_odd_ratio.txt')
    @Return   : 
                output --> output dataframe contain the odd ratio
    @Flow     : step1 --> retrieve the codon odd ratio
    '''

    # make the dataframe
    odd_ratio_df = pd.DataFrame()

    # for each odd_ratio file
    for odd_ratio_file in odd_ratio_list:
        # import the odd_ratio file
        odd_ratio = pd.read_csv(odd_ratio_file, sep='\t', index_col=False)

        # merge the gene odd_ratio file
        if odd_ratio_df.empty:
            odd_ratio_df = odd_ratio
        else:
            # delete the columns ['AA', 'Abbr.']
            odd_ratio_df = pd.concat([odd_ratio_df, odd_ratio], axis=0)

    return odd_ratio_df


def output_table(odd_ratio_merge, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the codon odd ratio
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    odd_ratio_merge.to_csv(output_file + '_sum_codon_odd_ratio.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_odd_ratio_args_parser()

    print('Step2: import the codon odd ratio file.', flush=True)
    odd_ratio_df = import_odd_ratio_files(args.list)

    print('Step3: output the codon odd ratio table.', flush=True)
    output_table(odd_ratio_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : SeRP.py


import os
import sys
from collections import OrderedDict

import matplotlib as mpl
import matplotlib.pyplot as plt

import numpy as np
import pandas as pd
import polars as pl

from Bio.Seq import Seq

from scipy.signal import savgol_filter
from scipy.stats import ranksums
from scipy.stats import ttest_rel


class SeRP(object):
    def __init__(self, args):
        self.tmp = None

        # input files
        self.rpf_file = args.rpf
        self.total_rpf = args.norm
        self.anno = args.anno
        self.gene_list = []
        self.gene_dict = OrderedDict()

        # parameters for analysis
        self.ck = args.control
        self.ip = args.ip
        self.ck_name = self.ck.split(',')
        self.ip_name = self.ip.split(',')

        # parameters for mrna import
        self.gene_list = None
        self.title = ['transcripts', 'gene_name']

        # parameters for rpf import
        self.all_rpf = pd.DataFrame()
        self.rpf_seq = pd.DataFrame()
        self.all_rpf_dict = OrderedDict()
        self.rpf_gene_list = pd.DataFrame()
        self.total_rpf_num = pd.DataFrame()
        self.mean_aa_rpm = None
        self.peak_rpm = []

        # parameters for rpf normalise
        self.background = args.background
        self.scale = args.scale
        self.fill = args.fill
        self.correlation = args.corr
        self.threshold = args.min

        # parameters for enrichment filter
        self.size = args.size
        self.backFold = args.backFold
        self.enrich = args.enrich
        self.gaps = args.gaps
        self.proportion = args.proportion
        self.width = args.width
        self.collision = args.collision

        self.upstream = args.upstream
        self.downstream = args.downstream

        # output peak detection files
        self.all = args.all
        self.peak_out = args.output
        self.ratio_out = args.ratio
        
        self.peak_log_file = self.peak_out + '_peaks.log'
        self.peak_txt_file = self.peak_out + '_peaks.txt'
        self.peak_bed_file = self.peak_out + '_peaks.bed'
        self.peak_sig_bed_file = self.peak_out + '_sig_peaks.bed'
        self.peak_seq_file = self.peak_out + '_peaks_sequence.txt'
        # self.peak_rpm_file = self.peak_out + '_peaks_rpm.txt'
        self.peak_ratio_file = self.peak_out + '_peaks_ratio.txt'
        self.enrich_ratio_file = self.peak_out + '_enrich_ratio.txt'

        # self.peak_merge = pd.DataFrame()
        self.peak_merge = []
        self.bed_merge = pd.DataFrame()

        # ratio
        self.all_ratio = OrderedDict()
        self.ave_ratio = OrderedDict()

        # output scripts and figure
        self.figure_out = args.fig
        self.enrich_out = []
        self.matlab_merge = OrderedDict()
        self.fig_merge = OrderedDict()
        self.peak_scripts_file = self.peak_out + '_peaks_scripts.m'
        self.none_peak_scripts_file = self.peak_out + '_peaks_none_scripts.m'

    # check all the arguments
    def args_check(self):
        # check the input file
        if not os.path.exists(self.rpf_file):
            print('ERROR! The specified Input file does not exist!', flush=True)
            sys.exit()

        # check the total rpf file
        if self.total_rpf and not os.path.exists(self.total_rpf):
            print('ERROR! The specified Total RPFs file does not exist!', flush=True)
            sys.exit()

        # check the gene annotation file
        if self.anno and not os.path.exists(self.anno):
            print('ERROR! The specified gene annotation file does not exist!', flush=True)
            sys.exit()

        # check the output file
        if os.path.exists(self.peak_log_file):
            print('Warning. The output file Peak_log exists and will be overwritten.', flush=True)
            os.remove(self.peak_log_file)

        if os.path.exists(self.peak_txt_file):
            print('Warning. The output file Peak_Annotation exists and will be overwritten.', flush=True)
            os.remove(self.peak_txt_file)

        if os.path.exists(self.peak_scripts_file):
            print('Warning. The output file of Peak_Scripts exists and will be overwritten.', flush=True)
            os.remove(self.peak_scripts_file)

        if os.path.exists(self.none_peak_scripts_file):
            print('Warning. The output file of None_Peak_Scripts exists and will be overwritten.', flush=True)
            os.remove(self.none_peak_scripts_file)

        if os.path.exists(self.peak_bed_file):
            print('Warning. The output file of Bed_Annotation exists and will be overwritten.', flush=True)
            os.remove(self.peak_bed_file)

        if os.path.exists(self.peak_ratio_file):
            print('Warning. The output file of Peak_Ratio exists and will be overwritten.', flush=True)
            os.remove(self.peak_ratio_file)

        # check the figure output path
        if self.figure_out:
            figure_out_path = self.peak_out + '_figures'
            if not os.path.exists(figure_out_path):
                os.makedirs(figure_out_path)
            else:
                print('Warning. The output directory already exists, and the figure will be output to '
                      'the existing directory.', flush=True)

    # function for gene annotation txt import
    def gene_anno(self):

        # if the annotation file is provided
        if self.anno:
            with open(self.anno, 'r') as anno_in:
                for line in anno_in:
                    rec = line.strip().split('\t')
                    self.gene_dict[rec[2]] = rec[1]
                # add the mrna not in the annotation file
                for mrna in self.gene_list:
                    if self.gene_dict.get(mrna):
                        pass
                    else:
                        self.gene_dict[mrna] = '-'
        else:
            for mrna in self.gene_list:
                self.gene_dict[mrna] = '-'

    # function for rpf txt import
    def rpf_txt_read(self):
        # get the input the samples name
        # samples = ck.split(',') + ip.split(',')
        # sample_num = len(ck_name + ip_name)
        col_index = ['name', 'from_tis', 'region']

        # import the data with polars
        all_rpf = pl.read_csv(self.rpf_file, separator='\t')
        # all_rpf = pd.read_csv(self.rpf_file, sep='\t', header=0, names=None)
        all_rpf = all_rpf.to_pandas()

        self.rpf_seq = all_rpf.iloc[:, 0:6]
        self.all_rpf = all_rpf.loc[:, col_index]

        for now_sp in self.ck_name + self.ip_name:
            now_sp_index = [now_sp + '_f0', now_sp + '_f1', now_sp + '_f2']
            self.all_rpf[now_sp] = all_rpf.loc[:, now_sp_index].sum(axis=1)

        col_title = self.all_rpf.columns
        self.gene_list = self.all_rpf.name.unique().tolist()
        rpf_group_gene = self.all_rpf.groupby('name')

        # Create the gene library.
        print('Create the gene library.', flush=True)
        # sys.stdout.flush()
        for mrna, grouped in rpf_group_gene:
            self.all_rpf_dict[mrna] = grouped

        # for mrna in self.gene_list:
        #     self.all_rpf_dict[mrna] = {}

        # with open(self.rpf_file, 'r') as cover_in:
        #     for line in islice(cover_in, 2, None):
        #         rec = line.strip().split('\t')
        #         self.all_rpf_dict[rec[0]][int(rec[3])] = rec
        #
        #     for mrna, mess in self.all_rpf_dict.items():
        #         self.all_rpf_dict[mrna] = pd.DataFrame.from_dict(mess, orient='index', columns=col_title)

        # import the total_rpf txt
        # the total rpf num for the RPM normalization
        if self.total_rpf:
            self.total_rpf_num = pd.read_csv(self.total_rpf, sep='\t', header=None, names=None, index_col=0)
            self.total_rpf_num = pd.Series(self.total_rpf_num[1], index=self.ck_name + self.ip_name)
            self.total_rpf_num = self.total_rpf_num.astype(int)
        else:
            self.total_rpf_num = self.all_rpf[self.ck_name + self.ip_name].sum()

        print('Total RPFs number:', flush=True)
        print(self.total_rpf_num, flush=True)

        # Calculate the mean value of CDS rpm for all genes and use it for background.
        ck_cds_rpf = pd.DataFrame(self.all_rpf[self.all_rpf.region == 'cds'], columns=col_index + self.ck_name)
        ck_cds_rpm = ck_cds_rpf[self.ck_name].astype('float').div(self.total_rpf_num.loc[self.ck_name].values).div(self.gene_list.__len__()) * self.scale
        self.mean_aa_rpm = ck_cds_rpm.sum() / self.background

    # define the format of output files
    def format_title(self):
        self.title = ['transcripts', 'gene_name']
        rpf_names = self.ck_name + self.ip_name
        rpm_names = [i + '_rpm' for i in rpf_names]
        results_title = ['min_corr_ck', 'min_corr_ip', 'comments', 'peak_num', 'peak_length', 'gaps', 'collision_start',
                         'peak_start', 'peak_end', 'collision_end', 'max_site', 'max_fold', 'mean_fold', 'ck_peak_rpm',
                         'ip_peak_rpm', 'P_Value', 'BHFDR', 'Peak_loci']
        self.title.extend(rpf_names + rpm_names + results_title)

    @staticmethod
    def safe_to_numeric(series):
        try:
            return pd.to_numeric(series)
        except ValueError:
            # ignore the error, and return the original series
            return series
    
    # define the format of rpf output list,
    def make_rpf_list(self, gene_rpf):
        gene_rpf_sum = gene_rpf[self.ck_name + self.ip_name].sum().tolist()
        # gene_rpf_mean = gene_rpf[ck_name + ip_name].mean().tolist()
        raw_gene_rpm = gene_rpf[self.ck_name + self.ip_name].div(self.total_rpf_num.values) * self.scale
        raw_gene_rpm_sum = raw_gene_rpm[self.ck_name + self.ip_name].sum().tolist()

        return gene_rpf_sum, raw_gene_rpm, raw_gene_rpm_sum

    # define the function for different protein binding modes.
    def flt_background_max_value(self, ip_name, raw_gene_rpm, cds_region):
        # condition 1: bound to ribosome
        if self.background == 0:
            return True
        # condition 2: bound to nascent polypeptides
        elif self.background == 30:

            max_back_value = raw_gene_rpm.loc[0:self.background - 1].max()
            # Exclude outliers of stop codon, "cds_region[-1]-1"
            max_bound_value = raw_gene_rpm.loc[self.background:cds_region[-1] - 1].max()

            # filter the strong binding score
            if self.backFold:
                delta_value = max_bound_value - max_back_value * self.enrich
            else:
                delta_value = max_bound_value - max_back_value * self.collision

            # flag-ip samples after background need 2 fold change higher than mock-ip
            if any(delta_value[ip_name] < 0):
                return False
            else:
                return True

    # define the function for different protein binding modes.
    def flt_background_mean_value(self, ip_name, raw_gene_rpm, cds_region):
        # condition 1: bound to ribosome
        if self.background == 0:
            return True
        
        # condition 2: bound to nascent polypeptides
        elif self.background == 30:
            
            mean_back_value = raw_gene_rpm.loc[0:self.background - 1].mean()
            # Exclude outliers of stop codon, "cds_region[-1]-1"
            mean_bound_value = raw_gene_rpm.loc[self.background:cds_region[-1] - 1].mean()

            # filter the strong binding score
            if self.backFold:
                delta_value = mean_bound_value - mean_back_value * self.enrich
            else:
                delta_value = mean_bound_value - mean_back_value * self.collision

            # flag-ip samples after background need 2 fold change higher than mock-ip
            if any(delta_value[ip_name] < 0):
                return False
            else:
                return True
            
    # define the function to calculate the correlation between the replicates
    def corr_calc(self, sp_gene_rpf_corr, sp_name):
        sp_corr_pairs = 0
        # more than one sample
        if len(sp_name) > 1:
            for rows in range(len(sp_name) - 1):
                for cols in range(rows, len(sp_name) - 1):
                    if sp_gene_rpf_corr.iloc[rows, cols + 1] >= self.correlation:
                        sp_corr_pairs += 1
        # only one sample in the group
        else:
            sp_corr_pairs = 1

        total_pairs = 0
        for counter in range(len(sp_name)):
            total_pairs += counter

        return sp_corr_pairs, total_pairs

    # define the function to fill the missing values with mean RPF
    def fill_empty(self, ck_name, raw_gene_rpm, cds_region):
        gene_rpm = raw_gene_rpm.copy()
        gene_rpm_mean = gene_rpm.loc[cds_region].mean()

        for sample in ck_name:
            # condition 1: fill the missing values with mean RPFs of total genes
            if self.fill == 0:
                # gene_rpm[sample].loc[gene_rpm[sample] == 0] = self.mean_aa_rpm[sample]
                gene_rpm.loc[gene_rpm[sample] == 0, sample] = self.mean_aa_rpm[sample]

            # condition 2: fill the missing values with mean RPFs of currently gene cds
            elif self.fill == -1:
                # gene_rpm[sample].loc[gene_rpm[sample] == 0] = gene_rpm_mean[sample]
                gene_rpm.loc[gene_rpm[sample] == 0, sample] = gene_rpm_mean[sample]

            # condition 3: fill the missing values with mean RPFs of specific region
            else:
                # Get the background mean value of the specific region
                gene_background = gene_rpm[sample].loc[0:self.fill - 1]
                counter_aa = len(gene_background[gene_background == 0].index)
                gene_background_mean = gene_background.mean()
                # If two-thirds of the specific region is covered by RPF.
                if counter_aa >= self.fill * 2 / 3:
                    # gene_rpm[sample].loc[gene_rpm[sample] == 0] = gene_background_mean
                    gene_rpm.loc[gene_rpm[sample] == 0, sample] = gene_background_mean

                # If half of the specific region is covered by RPF, and gene_rpm_mean <= gene_background_mean
                elif counter_aa >= self.fill / 2 and gene_rpm_mean[sample] <= gene_background_mean:
                    # gene_rpm[sample].loc[gene_rpm[sample] == 0] = gene_background_mean
                    gene_rpm.loc[gene_rpm[sample] == 0, sample] = gene_background_mean

                # if the specific region is noisy, use gene_rpm_mean instead
                elif counter_aa >= self.fill / 2 and gene_rpm_mean[sample] > gene_background_mean:
                    # gene_rpm[sample].loc[gene_rpm[sample] == 0] = gene_rpm_mean[sample]
                    gene_rpm.loc[gene_rpm[sample] == 0, sample] = gene_rpm_mean[sample]

                # if the specific region is noisy, use gene_rpm_mean instead
                elif counter_aa < self.fill / 2:
                    # gene_rpm[sample].loc[gene_rpm[sample] == 0] = gene_rpm_mean[sample]
                    gene_rpm.loc[gene_rpm[sample] == 0, sample] = gene_rpm_mean[sample]

                else:
                    # gene_rpm[sample].loc[gene_rpm[sample] == 0] = self.mean_aa_rpm[sample]
                    gene_rpm.loc[gene_rpm[sample] == 0, sample] = self.mean_aa_rpm[sample]

        return gene_rpm

    # define the function to calculate the enrichment ratio for each sample
    @staticmethod
    def calc_ratio(gene_rpm, ck_name, ip_name):
        gene_ratio = pd.DataFrame()
        # Ratio calculation is performed on any two samples in the two grouped data.
        for ip_sample in ip_name:
            for ck_sample in ck_name:
                gene_ratio[ip_sample + '_' + ck_sample] = gene_rpm[ip_sample].div(gene_rpm[ck_sample])

        return gene_ratio

    # define the function to smooth the data
    def smooth_data(self, gene_ratio, gene_rpf):
        # gene_ratio = gene_ratio.replace(np.inf, np.nan)
        gene_ratio_mean = gene_ratio.mean(axis=1)
        gene_ratio_mean.index = gene_rpf.from_tis

        if self.size == 0:
            gene_ratio_smooth = pd.DataFrame(gene_ratio_mean, index=gene_rpf.from_tis, columns=['enrich'])
            return gene_ratio_mean, gene_ratio_smooth
        else:
            gene_ratio_smooth = savgol_filter(gene_ratio_mean, self.size, 1, mode='nearest')
            gene_ratio_smooth = pd.DataFrame(gene_ratio_smooth, index=gene_rpf.from_tis, columns=['enrich'])
            return gene_ratio_mean, gene_ratio_smooth

    # define the function to get the peak annotation
    @staticmethod
    def peak_anno(utr5_region, cds_region, utr3_region, peak_start, peak_end):
        # check the peak start site
        peak_start_loci = ''
        if peak_start in utr5_region:
            peak_start_loci = 'utr5'
        elif peak_start in cds_region:
            peak_start_loci = 'cds'
        elif peak_start in utr3_region:
            peak_start_loci = 'utr3'
        # check the peak end site
        peak_end_loci = ''
        if peak_end in utr5_region:
            peak_end_loci = 'utr5'
        elif peak_end in cds_region:
            peak_end_loci = 'cds'
        elif peak_end in utr3_region:
            peak_end_loci = 'utr3'

        if peak_start_loci == peak_end_loci:
            return peak_start_loci
        else:
            return peak_start_loci + ',' + peak_end_loci

    # define the function to retrieve the collision region
    def retrieve_collision(self, gene_ratio_smooth, peak):
        gene_index = gene_ratio_smooth.index.tolist()
        peak_left_start, peak_left_end, peak_right_start, peak_right_end = [], [], [], []

        for peak_nums, peak_info in peak.items():
            peak_left_num = 0
            peak_right_num = 0

            peak_regions = list(peak_info.values())[0]
            # check the start position and gene start site
            start_posi = peak_regions[0] - gene_index[0]
            if start_posi == 0:
                pass
            else:
                left_shift_aa = 0
                for aa_num in range(start_posi):
                    left_shift_aa += 1
                    shift_aa_enrich = gene_ratio_smooth.loc[peak_regions[0] - aa_num - 1].enrich
                    if left_shift_aa >= 10:
                        break
                    elif shift_aa_enrich < self.collision:
                        break
                    elif shift_aa_enrich >= self.collision:
                        peak_left_num += 1

            # check the stop position and gene stop site
            stop_posi = gene_index[-1] - peak_regions[-1]
            if stop_posi == 0:
                pass
            else:
                right_shift_aa = 0
                for aa_num in range(stop_posi):
                    right_shift_aa += 1
                    shift_aa_enrich = gene_ratio_smooth.loc[peak_regions[-1] + aa_num + 1].enrich
                    if right_shift_aa >= 10:
                        break
                    elif shift_aa_enrich < self.collision:
                        break
                    elif shift_aa_enrich >= self.collision:
                        peak_right_num += 1

            peak_left_start.append(peak_regions[0] - peak_left_num)
            # peak_left_end.append(peak_regions[0])
            # peak_right_start.append(peak_regions[-1])
            peak_right_end.append(peak_regions[-1] + peak_right_num)

        return peak_left_start, peak_right_end

    # define the function to retrieve the maximum fold and position
    @staticmethod
    def get_max_enrich(peak, gene_ratio_smooth):
        max_peak_enrich = []
        max_peak_site = []
        mean_peak_enrich = []

        for peak_num, peak_dict in peak.items():
            peak_list = list(peak_dict.values())[0]
            gene_ratio_list = gene_ratio_smooth.loc[peak_list]

            # max_enrich = round(gene_ratio_list.max()[0], 4)
            max_enrich = round(gene_ratio_list.max().iloc[0], 4)
            max_peak_enrich.append(max_enrich)

            # max_site = str(gene_ratio_list.idxmax()[0])
            max_site = str(gene_ratio_list.idxmax().iloc[0])

            max_peak_site.append(max_site)

            mean_enrich = str(round(gene_ratio_list.mean().enrich, 4))
            mean_peak_enrich.append(mean_enrich)

        return max_peak_enrich, max_peak_site, mean_peak_enrich

    # define the function to calculate the PValue with Wilcoxon-test (rank sum test)
    @staticmethod
    def calc_rank_sum_v1(posi_list, raw_gene_rpm, ck_name, ip_name):
        rank_sum_results = []
        peak_range = posi_list
        ck_gene_rpf = raw_gene_rpm.loc[peak_range][ck_name].mean(axis=1)
        ip_gene_rpf = raw_gene_rpm.loc[peak_range][ip_name].mean(axis=1)
        t, p = ranksums(ck_gene_rpf, ip_gene_rpf)
        rank_sum_results.append(str(p))

        return rank_sum_results

    # define the function to calculate the PValue with Wilcoxon-test (rank sum test)
    @staticmethod
    def calc_rank_sum_v2(peak, raw_gene_rpm, ck_name, ip_name, mrna):
        rank_sum_results = []
        ck_rpm = []
        ip_rpm = []
        rpm_dict = OrderedDict()
        for peak_num, peak_dict in peak.items():
            peak_range = list(peak_dict.values())[0]
            ck_gene_rpf = raw_gene_rpm.loc[peak_range][ck_name].mean(axis=1)
            ip_gene_rpf = raw_gene_rpm.loc[peak_range][ip_name].mean(axis=1)
            t, p = ranksums(ck_gene_rpf, ip_gene_rpf)
            rank_sum_results.append(str(p))
            ck_rpm.append(ck_gene_rpf.sum())
            ip_rpm.append(ip_gene_rpf.sum())

            names = mrna + '_peak_' + str(peak_num)
            rpm_dict[names] = list(map(str, raw_gene_rpm.loc[peak_range][ck_name + ip_name].sum(axis=0).to_list()))

        return rank_sum_results, ck_rpm, ip_rpm, rpm_dict

    # define the function to calculate the PValue with paired t-test
    @staticmethod
    def calc_ttest_v1(posi_list, raw_gene_rpm, ck_name, ip_name):
        ttest_results = []
        peak_range = posi_list
        ck_gene_rpf = raw_gene_rpm.loc[peak_range][ck_name].mean(axis=1)
        ip_gene_rpf = raw_gene_rpm.loc[peak_range][ip_name].mean(axis=1)
        t, p = ttest_rel(ck_gene_rpf, ip_gene_rpf)
        ttest_results.append(str(p))

        return ttest_results

    # define the function to calculate the PValue with Wilcoxon-test (rank sum test)
    @staticmethod
    def calc_ttest_v2(peak, raw_gene_rpm, ck_name, ip_name, mrna):
        ttest_results = []
        ck_rpm = []
        ip_rpm = []
        rpm_dict = OrderedDict()
        for peak_num, peak_dict in peak.items():
            peak_range = list(peak_dict.values())[0]
            ck_gene_rpf = raw_gene_rpm.loc[peak_range][ck_name].mean(axis=1)
            ip_gene_rpf = raw_gene_rpm.loc[peak_range][ip_name].mean(axis=1)
            t, p = ttest_rel(ck_gene_rpf, ip_gene_rpf)
            ttest_results.append(str(p))
            ck_rpm.append(ck_gene_rpf.sum())
            ip_rpm.append(ip_gene_rpf.sum())

            names = mrna + '_peak_' + str(peak_num)
            rpm_dict[names] = list(map(str, raw_gene_rpm.loc[peak_range][ck_name + ip_name].sum(axis=0).to_list()))

        return ttest_results, ck_rpm, ip_rpm, rpm_dict

    # define the function to find the peak
    def find_peak(self, eligible_index, gene_ratio_smooth, raw_gene_rpm, ck_name, ip_name):
        peak = OrderedDict()
        pre_peak = OrderedDict()
        temp_store = []
        pre_peak_num = 1
        now_length = 0

        # if none gaps allowed
        if self.gaps == 0:
            for posi in range(len(eligible_index)):
                if not temp_store:
                    temp_store.append(eligible_index[posi])
                    now_length = 1
                else:
                    now_gap_length = eligible_index[posi] - eligible_index[posi - 1] - 1
                    # Check for gaps inside the potential peak.
                    if now_gap_length == 0:
                        temp_store.append(eligible_index[posi])
                        now_length += 1
                    else:
                        if now_length >= self.width:
                            pre_peak[pre_peak_num] = temp_store
                            pre_peak_num += 1
                            temp_store = [eligible_index[posi]]
                            now_length = 1

                        elif now_length < self.width:
                            temp_store = [eligible_index[posi]]
                            now_length = 1
            # save the last peak
            if now_length >= self.width:
                pre_peak[pre_peak_num] = temp_store

            # Traverse each potential peak.
            now_peak_num = 0
            for peak_num, peak_posi in pre_peak.items():
                # [peak length, peak position]
                peak[now_peak_num] = [peak_posi[-1] - peak_posi[0] + 1, peak_posi]
                now_peak_num += 1

        else:
            # This step used to search the potential peak but do not combine them
            # Traverse all potential peak regions and use the gaps_threshold for segmentation.
            for posi in range(len(eligible_index)):
                if not temp_store:
                    temp_store.append(eligible_index[posi])
                    now_length = 1
                else:
                    now_gap_length = eligible_index[posi] - eligible_index[posi - 1] - 1

                    # Check for gaps inside the potential peak.
                    if now_gap_length == 0:
                        temp_store.append(eligible_index[posi])
                        now_length += 1

                    # The loop skips when the enrichment in the gap lower than Collision.
                    elif now_gap_length <= self.gaps:
                        # if the gap enrich high than specified collision ratio
                        gap_fold_ratio = gene_ratio_smooth.loc[eligible_index[posi - 1] + 1:eligible_index[posi] - 1] - self.collision
                        # The loop continues when the enrichment in the gap higher than Collision.
                        if gap_fold_ratio.values.min() >= 0:
                            temp_store.append(eligible_index[posi])
                            now_length = now_length + now_gap_length + 1

                        # Save the results and break the current loop,
                        # when the enrichment in the gap lower than Collision and the
                        # current peak width is wider than the peak_width.
                        elif gap_fold_ratio.values.min() < 0 and now_length >= self.width:
                            pre_peak[pre_peak_num] = temp_store
                            pre_peak_num += 1
                            temp_store = [eligible_index[posi]]
                            now_length = 1

                        else:
                            temp_store = [eligible_index[posi]]
                            now_length = 1

                    # If the length of the consecutive gap is greater than the gap_threshold
                    elif now_gap_length > self.gaps:
                        if now_length >= self.width:
                            pre_peak[pre_peak_num] = temp_store
                            pre_peak_num += 1
                            temp_store = [eligible_index[posi]]
                            now_length = 1

                        elif now_length < self.width:
                            temp_store = [eligible_index[posi]]
                            now_length = 1

            if now_length >= self.width:
                pre_peak[pre_peak_num] = temp_store

            # Traverse each potential peak.
            now_peak_num = 0
            for peak_num, peak_posi in pre_peak.items():
                # if the current peak meets the conditions.
                if len(peak_posi) / (peak_posi[-1] - peak_posi[0] + 1) >= 1 - self.proportion:
                    peak[now_peak_num] = [peak_posi[-1] - peak_posi[0] + 1, peak_posi]
                    now_peak_num += 1
                # if the current peak does not meets the conditions.
                else:

                    temp_split_list = []  # split the pre peak region by the gap
                    temp_list = []
                    now_region_num = 0

                    # Split the remaining candidate peaks again.
                    for posi in range(len(peak_posi)):
                        if not temp_split_list:
                            temp_split_list.append(peak_posi[posi])
                        # If the gaps does not exist
                        elif peak_posi[posi] - peak_posi[posi - 1] == 1:
                            temp_split_list.append(peak_posi[posi])
                        # If the gaps exist
                        elif peak_posi[posi] - peak_posi[posi - 1] > 1:
                            temp_list.append(temp_split_list)
                            temp_split_list = [peak_posi[posi]]
                            now_region_num += 1

                    if temp_split_list:
                        temp_list.append(temp_split_list)

                    # Regroup the divided candidate peaks in order.
                    # The loop continues when there is still candidate peaks in pre_peak Dict.
                    region_num = len(temp_list)
                    permutations = OrderedDict()
                    for start_posi in range(region_num - 1):
                        # Exclude peaks whose width is less than 5 AA.
                        if len(temp_list[start_posi]) >= self.width:
                            if len(temp_list[start_posi]) / (
                                    temp_list[start_posi][0] - temp_list[start_posi][-1] + 1) >= 1 - self.proportion:
                                ttest_results = self.calc_ttest_v1(temp_list[start_posi], raw_gene_rpm, ck_name,
                                                                         ip_name)
                                permutations[str(start_posi)] = [ttest_results, len(temp_list[start_posi]), start_posi,
                                                                 temp_list[start_posi]]
                            else:
                                pass
                        for stop_posi in range(start_posi + 1, region_num):
                            merge_region = []
                            for i in temp_list[start_posi:stop_posi + 1]:
                                merge_region.extend(i)

                            merge_len = int(merge_region[-1]) - int(merge_region[0]) + 1
                            # If the current permutation meets the conditions.
                            if len(merge_region) / merge_len >= 1 - self.proportion:
                                permutations_num = '_'.join([str(num) for num in range(start_posi, stop_posi + 1)])
                                ttest_results = self.calc_ttest_v1(temp_list[start_posi], raw_gene_rpm, ck_name,
                                                                         ip_name)
                                permutations[permutations_num] = [ttest_results, merge_len, permutations_num,
                                                                  merge_region]
                            else:
                                pass

                    # Filter out the best results from different permutations.
                    while len(permutations.keys()) >= 1:
                        # If there is only one element in the list.
                        if len(permutations.keys()) == 1:
                            peak_info = list(permutations.values())[0]
                            peak[now_peak_num] = [peak_info[1], peak_info[-1]]
                            break
                        else:
                            # If there are multiple elements in the list.
                            result = [i for i in permutations.keys() if '_' in i]

                            # filter the optimal permutation
                            if result and not(self.all):
                                now_best_permutations = []
                                for permutations_num, permutations_info in permutations.items():
                                    if len(now_best_permutations) == 0:
                                        now_best_permutations = permutations_info

                                    # if the length of current permutation large than the optimal one.
                                    elif permutations_info[1] > now_best_permutations[1]:
                                        # compare the p_value and gap proportion of the current permutation and the optimal one.
                                        gap_per = (int(permutations_info[3][-1]) - int(permutations_info[3][0])) / len(
                                            permutations_info[3])
                                        now_best_gap_per = (int(permutations_info[3][-1]) - int(
                                            permutations_info[3][0])) / len(permutations_info[3])
                                        if float(permutations_info[0][0]) > float(now_best_permutations[0][0]) and gap_per > now_best_gap_per:
                                            continue
                                        else:
                                            now_best_permutations = permutations_info

                                    # if the length of current permutation equal to the optimal one.
                                    elif permutations_info[1] == now_best_permutations[1]:
                                        if float(permutations_info[0][0]) < float(now_best_permutations[0][0]):
                                            now_best_permutations = permutations_info

                                        # Compare the p_value of the current permutation and the optimal one.
                                        elif float(permutations_info[0][0]) == float(now_best_permutations[0][0]):
                                            # compare the gap proportion of the current permutation and the optimal one.
                                            gap_per = (int(permutations_info[3][-1]) - int(permutations_info[3][0]))/len(permutations_info[3])
                                            now_best_gap_per = (int(permutations_info[3][-1]) - int(permutations_info[3][0]))/len(permutations_info[3])
                                            if gap_per < now_best_gap_per:
                                                now_best_permutations = permutations_info
                                            elif gap_per > now_best_gap_per:
                                                continue

                                            # Compare the start position of the current permutation and the optimal one.
                                            elif int(permutations_num.split('_')[0]) < int(now_best_permutations[2].split('_')[0]):
                                                now_best_permutations = permutations_info

                                        elif float(permutations_info[0][0]) > float(now_best_permutations[0][0]):
                                            continue
                                        else:
                                            # now_best_permutations
                                            sys.stdout.write('unknown')

                                    # If the width of the current permutation is narrower than the optimal one.
                                    elif permutations_info[1] < now_best_permutations[1]:
                                        continue

                                # save the best permutations
                                peak_info = now_best_permutations[-1]
                                peak[now_peak_num] = [int(peak_info[-1]) - int(peak_info[0]) + 1, peak_info]
                                now_peak_num += 1
                                permutations.pop(now_best_permutations[2])

                                # delete the permutations and original elements
                                permutations_elements = now_best_permutations[2].split('_')
                                shift_posi = 0
                                for start_posi in permutations_elements:
                                    shift_posi += 1
                                    if permutations.get(start_posi):
                                        permutations.pop(start_posi)
                                    for stop_posi in permutations_elements[shift_posi:]:
                                        permutations_num = '_'.join([str(num) for num in range(int(start_posi), int(stop_posi) + 1)])
                                        if permutations.get(permutations_num):
                                            permutations.pop(permutations_num)
                            # If the elements in the list are all single
                            else:
                                for permutations_num, permutations_info in permutations.items():
                                    peak[now_peak_num] = [len(permutations_info[-1]), permutations_info[-1]]
                                    now_peak_num += 1
                                break

        # sort the peak by region
        peak_sort = OrderedDict()
        if not bool(peak):
            return peak_sort
        elif len(peak.keys()) == 1:
            peak_sort[0] = {peak[0][0]: peak[0][1]}
            return peak_sort
        else:
            peak_sort1 = sorted(peak.items(), key=lambda d: d[1][1][0])
            new_peak_num = 0
            for peak_info in peak_sort1:
                peak_sort[new_peak_num] = {peak_info[1][0]: peak_info[1][1]}
                new_peak_num += 1

            return peak_sort

    # define the function to save the matlab scripts for none peak results
    def matlab_none_peak_scripts(self, scripts_out, mrna, gene_ratio_smooth, cds_region):
        mrna_name = mrna.replace('-', '_').replace('.', '_')
        title_name = mrna_name.replace('_', '-')
        # ylim_down = gene_ratio_smooth.min().enrich - gene_ratio_smooth.min().enrich * 0.1
        # ylim_up = gene_ratio_smooth.max().enrich + gene_ratio_smooth.max().enrich * 0.1

        scripts_out.writelines(''.join("clear\n"))
        scripts_out.writelines(' '.join([mrna_name, '= [\n']))
        scripts_out.writelines(' '.join([str(i) for i in gene_ratio_smooth.index.to_list()]) + '\n')
        scripts_out.writelines(' '.join([str(round(float(i[0]), 4)) for i in gene_ratio_smooth.values.tolist()]) + '\n];\n')
        scripts_out.writelines(''.join("figure('Visible','off');") + '\n')
        scripts_out.writelines(''.join("set(gcf, 'units', 'normalized', 'position', [0.1 0.1 0.4 0.3]);") + '\n')
        scripts_out.writelines(''.join("hold on") + '\n')
        scripts_out.writelines(
            ''.join("plot(%s(1,:),%s(2,:),'Color','#0072bd','LineWidth', 0.8);" % (mrna_name, mrna_name)) + '\n')
        scripts_out.writelines(''.join("yline(%s,'Color','#999999','LineStyle','-.');" % self.enrich) + '\n')
        scripts_out.writelines(''.join("xline(%s,'Color','#999999','LineStyle','-.');" % cds_region[0]) + '\n')
        scripts_out.writelines(''.join("xline(%s,'Color','#999999','LineStyle','-.');" % cds_region[-1]) + '\n')
        scripts_out.writelines(''.join("xlim([%s(1,1) %s(1,end)]);" % (mrna_name, mrna_name)) + '\n')
        # scripts_out.writelines(''.join("%%ylim([%s %s]);" % (math.floor(ylim_down), math.ceil(ylim_up))) + '\n')
        scripts_out.writelines(''.join("box on;") + '\n')
        scripts_out.writelines(''.join("legend({'un-bound'}, 'location', 'EO');") + '\n')
        scripts_out.writelines(''.join("xlabel('AA position on mRNA');") + '\n')
        scripts_out.writelines(''.join("ylabel('Enrichment [a.u]');") + '\n')
        scripts_out.writelines(''.join("title('%s (%s)');" % (title_name, self.gene_dict[mrna])) + '\n')
        scripts_out.writelines(''.join("set(gca, 'FontName', 'Arial', 'FontSize', 12)") + '\n')
        scripts_out.writelines(''.join("hold off") + '\n')
        scripts_out.writelines(''.join("print('%s', '-r300', '-dpng');" % mrna) + '\n')
        scripts_out.writelines(''.join("saveas(gcf, 'pdf', '%s');" % mrna) + '\n\n')

    # define the function to save the matlab scripts for peak results
    def matlab_peak_scripts(self, scripts_out, mrna, gene_ratio_smooth, peak, peak_left_start, peak_right_end,
                            cds_region):
        mrna_name = mrna.replace('-', '_').replace('.', '_')
        title_name = mrna_name.replace('_', '-')
        collision_region = []
        peak_region = []
        collision_ratio = gene_ratio_smooth.copy()
        peak_ratio = gene_ratio_smooth.copy()
        # ylim_down = gene_ratio_smooth.min().enrich - abs(gene_ratio_smooth.min().enrich * 0.1)
        # ylim_up = gene_ratio_smooth.max().enrich + abs(gene_ratio_smooth.max().enrich * 0.1)

        for peak_num, peak_dict in peak.items():
            peak_len = list(peak_dict.keys())[0]
            peak_start = str(peak[peak_num][peak_len][0])
            peak_end = str(peak[peak_num][peak_len][-1])
            collision_start = peak_left_start[peak_num]
            collision_end = peak_right_end[peak_num]
            collision_region.extend([i for i in range(collision_start, int(peak_start) + 1)] +
                                    [j for j in range(int(peak_end), collision_end + 1)])
            peak_region.extend([i for i in range(int(peak_start), int(peak_end) + 1)])

        collision_ratio.loc[collision_ratio.index.difference(collision_region)] = np.nan
        peak_ratio.loc[collision_ratio.index.difference(peak_region)] = np.nan

        scripts_out.writelines(''.join("clear\n"))
        scripts_out.writelines(' '.join([mrna_name, '= [\n']))
        scripts_out.writelines(' '.join([str(i) for i in gene_ratio_smooth.index.to_list()]) + '\n')
        scripts_out.writelines(' '.join([str(round(float(i[0]), 4)) for i in gene_ratio_smooth.values.tolist()]) + '\n')
        scripts_out.writelines(' '.join([str(round(float(i[0]), 4)) for i in collision_ratio.values.tolist()]) + '\n')
        scripts_out.writelines(' '.join([str(round(float(i[0]), 4)) for i in peak_ratio.values.tolist()]) + '\n];\n')
        scripts_out.writelines(''.join("figure('Visible','off');") + '\n')
        scripts_out.writelines(''.join("set(gcf, 'units', 'normalized', 'position', [0.1 0.1 0.4 0.3]);") + '\n')
        scripts_out.writelines(''.join("hold on") + '\n')
        scripts_out.writelines(
            ''.join("plot(%s(1,:),%s(2,:),'Color','#0072bd','LineWidth', 0.8);" % (mrna_name, mrna_name)) + '\n')
        scripts_out.writelines(
            ''.join("plot(%s(1,:),%s(3,:),'Color','#edb120','LineWidth', 0.8);" % (mrna_name, mrna_name)) + '\n')
        scripts_out.writelines(
            ''.join("plot(%s(1,:),%s(4,:),'Color','#d95319','LineWidth', 0.8);" % (mrna_name, mrna_name)) + '\n')
        scripts_out.writelines(''.join("yline(%s,'Color','#999999','LineStyle','-.');" % self.enrich) + '\n')
        scripts_out.writelines(''.join("xline(%s,'Color','#999999','LineStyle','-.');" % cds_region[0]) + '\n')
        scripts_out.writelines(''.join("xline(%s,'Color','#999999','LineStyle','-.');" % cds_region[-1]) + '\n')
        scripts_out.writelines(''.join("xlim([%s(1,1) %s(1,end)]);" % (mrna_name, mrna_name)) + '\n')
        # scripts_out.writelines(''.join("%%ylim([%f %f]);" % (math.floor(ylim_down), math.ceil(ylim_up))) + '\n')
        scripts_out.writelines(''.join("box on;") + '\n')
        scripts_out.writelines(''.join("legend({'un-bound', 'edge', 'bound'}, 'location', 'EO');") + '\n')
        scripts_out.writelines(''.join("xlabel('AA position on mRNA');") + '\n')
        scripts_out.writelines(''.join("ylabel('Enrichment [a.u]');") + '\n')
        scripts_out.writelines(''.join("title('%s (%s)');" % (title_name, self.gene_dict[mrna])) + '\n')
        scripts_out.writelines(''.join("set(gca, 'FontName', 'Arial', 'FontSize', 10)") + '\n')
        scripts_out.writelines(''.join("hold off") + '\n')
        scripts_out.writelines(''.join("print('%s', '-r300', '-dpng');" % mrna) + '\n')
        scripts_out.writelines(''.join("save_pdf('%s');" % mrna) + '\n\n')

        return collision_ratio, peak_ratio

    # define the function to draw the figure
    def draw_figure(self, mrna, gene_ratio_smooth, collision_ratio, peak_ratio, cds_region):
        figure_out_path = './' + self.peak_out + '_figures'

        # draw the line plot
        mpl.rcParams['axes.linewidth'] = 0.64
        mpl.rcParams['xtick.labelsize'] = 6
        mpl.rcParams['ytick.labelsize'] = 6
        mpl.rcParams['legend.fontsize'] = 8

        fig = plt.figure(figsize=(5, 3), dpi=300)
        plt.subplots_adjust(0.1, 0.15, 0.9, 0.85)
        ax = fig.add_subplot(111)

        # gene_ratio.plot()
        plt.tick_params(axis='x', which='both',
                        bottom='on', top='off',
                        labelbottom='on',
                        direction='out', width=0.5, length=2.0)
        plt.tick_params(axis='y', which='both',
                        left='on', right='off',
                        labelleft='on', labelright='off',
                        direction='out', width=0.5, length=2.0)
        x = gene_ratio_smooth.index.tolist()
        plt.plot(x, gene_ratio_smooth.values.tolist(), color='#0072bd', linewidth=0.8, label='un-bound')
        plt.plot(x, collision_ratio.values.tolist(), color='#edb120', linewidth=0.8, label='bound')
        plt.plot(x, peak_ratio.values.tolist(), color='#d95319', linewidth=0.8, label='strongly-bound')
        plt.axvline(x=cds_region[0], c="#999999", ls="-.", lw=0.5)
        plt.axvline(x=cds_region[-1], c="#999999", ls="-.", lw=0.5)
        plt.axhline(y=self.enrich, c="#999999", ls="-.", lw=0.5)
        plt.legend()
        # plt.fill_between(x, gene_ratio.iloc[:, :].max(axis=1),
        #                 gene_ratio.iloc[:, :].min(axis=1), color='blue', linewidth=0.0, alpha=0.1)
        # plt.yscale('log', basey=2)
        ax.set_xlabel("AA position on mRNA", fontsize=8)
        ax.set_ylabel("Enrichment [a.u]", fontsize=8)
        ax.set_xlim(x[0], x[-1])
        ax.set_title(mrna + ' (' + self.gene_dict[mrna] + ')', fontsize=10)
        plt.savefig(figure_out_path + '/' + mrna + '.png')
        plt.close()

    # define the function to merge the output results
    def result_output(self, mrna, gene_rpf_sum, raw_gene_rpm_sum, ck_min_corr, ip_min_corr, peak,
                      peak_left_start, peak_right_end, max_peak_enrich,
                      max_peak_site, mean_peak_enrich, ck_rpm, ip_rpm, ttest_results, peak_results, utr5_region,
                      cds_region, utr3_region):
        
        for peak_num, peak_dict in peak.items():
            peak_len = list(peak_dict.keys())[0]
            peak_start = peak[peak_num][peak_len][0]
            peak_end = peak[peak_num][peak_len][-1]
            gaps = (peak_end - peak_start + 1) - peak_len
            # annotate the peak location in utr/cds
            peal_loci = self.peak_anno(utr5_region, cds_region, utr3_region, peak_start, peak_end)

            # save the peak results
            peak_tmp = '\t'.join([mrna, self.gene_dict[mrna], '\t'.join([str(i) for i in gene_rpf_sum]),
                                  '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                  ck_min_corr, ip_min_corr, 'Qualified', str(peak_num + 1),
                                  str(peak_len), str(gaps),
                                  str(peak_left_start[peak_num]),
                                  str(peak_start), str(peak_end),
                                  str(peak_right_end[peak_num]),
                                  str(max_peak_site[peak_num]),
                                  str(max_peak_enrich[peak_num]),
                                  mean_peak_enrich[peak_num],
                                  str(ck_rpm[peak_num]), str(ip_rpm[peak_num]),
                                  ttest_results[peak_num], '-',
                                  peal_loci])
            # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
            self.peak_merge.append(peak_tmp.split('\t'))

            peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')

        peak_results.flush()

    # define the function to detect binding peaks
    def detect_binding_peaks(self):
        # check and rename the output files
        peak_results = open(self.peak_log_file, 'w')

        # output the scripts of figures

        scripts_peak = open(self.peak_scripts_file, 'w')
        scripts_none_peak = open(self.none_peak_scripts_file, 'w')

        # get the samples name
        ck_name = self.ck.split(',')
        ip_name = self.ip.split(',')
        # col_index = ['transcript', 'from_tis', 'region']
        sample_num = len(ck_name + ip_name)

        # save the results
        self.format_title()
        peak_results.writelines('\t'.join(self.title) + '\n')

        # calc the ratio and find the peak
        for mrna in self.gene_list:
            # now gene is :
            print(mrna, flush=True)

            # get the gene rpf
            # gene_rpf = pd.DataFrame(self.all_rpf_dict[mrna], columns=col_index + self.ck_name + self.ip_name)
            gene_rpf = self.all_rpf_dict[mrna]
            # gene_rpf = gene_rpf.apply(pd.to_numeric, errors='ignore')
            gene_rpf = gene_rpf.apply(self.safe_to_numeric)
            gene_rpf.index = gene_rpf.from_tis

            # normalise the data to rpm
            gene_rpf_sum, raw_gene_rpm, raw_gene_rpm_sum = self.make_rpf_list(gene_rpf)

            # filter the rpf count >= threshold
            sp_rpf_threshold = 0
            for rpf_sum in gene_rpf_sum:
                if rpf_sum > self.threshold:
                    sp_rpf_threshold += 1

            if sp_rpf_threshold < sample_num:
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      '-', '-', 'Too few RPFs', '-', '-', '-', '-', '-', '-', '-',
                                      '-', '-', '-', '-', '-', '-', '-', '-'])
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')
                continue

            # get the utr and cds region，, for the peak region annotation
            utr5_region = gene_rpf[gene_rpf.region == '5utr'].index.tolist()
            cds_region = gene_rpf[gene_rpf.region == 'cds'].index.tolist()
            utr3_region = gene_rpf[gene_rpf.region == '3utr'].index.tolist()

            # set the threshold of the background region
            first_aa = self.flt_background_max_value(ip_name, raw_gene_rpm, cds_region)
            # first_aa = self.flt_background_mean_value(ip_name, raw_gene_rpm, cds_region)

            if first_aa:
                pass
            else:
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      '-', '-', 'Background value too large', '-', '-', '-', '-', '-', '-',
                                      '-', '-', '-', '-', '-', '-', '-', '-', '-'])
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')


                # use the mean_aa_rpm to fill the empty rpf.
                gene_rpm = self.fill_empty(ck_name, raw_gene_rpm, cds_region)

                # calculate the enrichment ratio for each sample
                gene_ratio = self.calc_ratio(gene_rpm, ck_name, ip_name)

                # smooth the data
                gene_ratio_mean, gene_ratio_smooth = self.smooth_data(gene_ratio, gene_rpf)

                # add the collision and peak ratio column with nan value
                enrich_out = pd.concat([gene_rpf, gene_ratio_smooth], axis=1).reset_index(drop=True)
                enrich_out = pd.concat([enrich_out, pd.DataFrame(np.nan, index=enrich_out.index, columns=['edge', 'bound'])], axis=1)

                self.enrich_out.append(enrich_out.reset_index(drop=True))

                continue

            # use the mean_aa_rpm to fill the empty rpf.
            gene_rpm = self.fill_empty(ck_name, raw_gene_rpm, cds_region)

            # calculate the enrichment ratio for each sample
            gene_ratio = self.calc_ratio(gene_rpm, ck_name, ip_name)

            # smooth the data
            gene_ratio_mean, gene_ratio_smooth = self.smooth_data(gene_ratio, gene_rpf)

            # combine the results of each gene
            self.all_ratio[mrna] = gene_ratio
            # ave_ratio[mrna] = gene_ratio_mean
            # cds_ratio[mrna] = gene_ratio.loc[cds_region]

            # calculate the correlation of samples
            ck_gene_rpf_corr = raw_gene_rpm[ck_name].corr()
            ck_min_corr = str(round(ck_gene_rpf_corr.values.min(), 4))
            
            ip_gene_rpf_corr = raw_gene_rpm[ip_name].corr()
            ip_min_corr = str(round(ip_gene_rpf_corr.values.min(), 4))

            # filter the correlation of the ck replicates： default 0.5
            ck_corr_pairs, total_pairs = self.corr_calc(ck_gene_rpf_corr, ck_name)

            if ck_corr_pairs < total_pairs:
                self.matlab_none_peak_scripts(scripts_none_peak, mrna, gene_ratio_smooth, cds_region)
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      ck_min_corr, ip_min_corr, 'Poor repeatability of ck samples',
                                      '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-',
                                      '-', '-'])
                
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')
                continue

            # filter the correlation of the ip replicates： default 0.5
            ip_corr_pairs, total_pairs = self.corr_calc(ip_gene_rpf_corr, ip_name)

            if ip_corr_pairs < total_pairs:
                self.matlab_none_peak_scripts(scripts_none_peak, mrna, gene_ratio_smooth, cds_region)
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      ck_min_corr, ip_min_corr, 'Poor repeatability of ip samples',
                                      '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-',
                                      '-', '-'])
                
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')
                continue

            # get the index of the rpm >= fold change,
            # can specify 1.2 for the collision and the 2 for the binding region
            eligible_index = gene_ratio_smooth[gene_ratio_smooth.enrich >= self.enrich].index.tolist()
            # index_12 = gene_ratio_smooth[gene_ratio_smooth.enrich >= 1.2].index.tolist()

            # filter the total aa number >= 5
            if len(eligible_index) < self.width:
                self.matlab_none_peak_scripts(scripts_none_peak, mrna, gene_ratio_smooth, cds_region)
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      ck_min_corr, ip_min_corr, 'Qualified',
                                      '0', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-',
                                      '-', '-'])
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')
                continue

            # find_peak main programme is here !!!
            # get the peaks from the smooth data
            peak = self.find_peak(eligible_index, gene_ratio_smooth, raw_gene_rpm, ck_name, ip_name)

            # if no peak in current gene
            if not bool(peak):
                self.matlab_none_peak_scripts(scripts_none_peak, mrna, gene_ratio_smooth, cds_region)
                peak_tmp = '\t'.join([mrna, self.gene_dict[mrna],
                                      '\t'.join([str(i) for i in gene_rpf_sum]),
                                      '\t'.join([str(round(i, 4)) for i in raw_gene_rpm_sum]),
                                      ck_min_corr, ip_min_corr, 'Qualified',
                                      '0', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-', '-',
                                      '-', '-'])
                # self.peak_merge = self.peak_merge.append(pd.Series(peak_tmp.split('\t')), ignore_index=True)
                self.peak_merge.append(peak_tmp.split('\t'))
                peak_results.writelines('\t'.join(peak_tmp.split('\t')) + '\n')

                enrich_out = pd.concat([gene_rpf, gene_ratio_smooth], axis=1).reset_index(drop=True)
                # add the collision and peak ratio column with nan value
                enrich_out = pd.concat([enrich_out, pd.DataFrame(np.nan, index=enrich_out.index, columns=['edge', 'bound'])], axis=1)

                self.enrich_out.append(enrich_out.reset_index(drop=True))

            else:
                # get the collision region from the smooth data
                peak_left_start, peak_right_end = self.retrieve_collision(gene_ratio_smooth, peak)

                # get the maximum site and fold change
                max_peak_enrich, max_peak_site, mean_peak_enrich = self.get_max_enrich(peak, gene_ratio_smooth)

                # calculate the p value of peaks by rank sum test
                ttest_results, ck_rpm, ip_rpm, rpm_dict = self.calc_ttest_v2(peak, raw_gene_rpm, ck_name, ip_name,
                                                                                   mrna)

                for peak_names, now_peak_rpm in rpm_dict.items():
                    mess = peak_names + '\t' + '\t'.join(now_peak_rpm) + '\n'
                    self.peak_rpm.append(mess)

                # output the peaks in txt format
                self.result_output(mrna, gene_rpf_sum, raw_gene_rpm_sum, ck_min_corr, ip_min_corr, peak,
                                   peak_left_start, peak_right_end,
                                   max_peak_enrich, max_peak_site, mean_peak_enrich, ck_rpm, ip_rpm, ttest_results,
                                   peak_results, utr5_region, cds_region, utr3_region)

                # output the figure scripts
                collision_ratio, peak_ratio = self.matlab_peak_scripts(scripts_peak, mrna, gene_ratio_smooth, peak,
                                                                       peak_left_start, peak_right_end, cds_region)

                # Output the demo figure
                if self.figure_out:
                    self.draw_figure(mrna, gene_ratio_smooth, collision_ratio, peak_ratio, cds_region)
                
                collision_ratio.rename(columns={'enrich': 'edge'}, inplace=True)
                peak_ratio.rename(columns={'enrich': 'bound'}, inplace=True)
                self.enrich_out.append(pd.concat([gene_rpf, gene_ratio_smooth, collision_ratio, peak_ratio], axis=1).reset_index(drop=True))

        self.peak_merge = pd.DataFrame(self.peak_merge)
        self.peak_merge.columns = self.title

        peak_results.close()
        scripts_none_peak.close()
        scripts_peak.close()

        print("transcripts number: %s" % len(self.gene_list), flush=True)

    # define the function to output the RPM of each peak
    def output_peak_rpm(self):
        # save the peak rpm
        split_line = pd.DataFrame(["##############################################"])
        peak_rpm = open(self.peak_rpm_file, 'w')
        # for mrna, ratio in self.all_ratio.items():
        #     ratio.index.name = mrna
        #     split_line.to_csv(self.peak_ratio_file, mode='a+', header=False, index=False)
        #     ratio.to_csv(self.peak_ratio_file, mode='a+', header=True, index=True)

    # define the function to output the original ratio data
    def output_ratio(self):
        split_line = pd.DataFrame(["##############################################"])
        for mrna, ratio in self.all_ratio.items():
            ratio.index.name = mrna
            split_line.to_csv(self.enrich_ratio_file, mode='a+', header=False, index=False)
            ratio.to_csv(self.enrich_ratio_file, mode='a+', sep = '\t', header=True, index=True)

    # define the function to output the peak message
    def output_peak(self):
        peak_all = self.peak_merge.loc[self.peak_merge.iloc[:, -3] != '-', ].copy()

        # peak_all.columns = self.title
        peak_all['P_Value'] = peak_all['P_Value'].astype('float')
        peak_all['rank'] = peak_all['P_Value'].rank(method='first')
        peak_num = peak_all['rank'].max()

        # calculate the FDR
        for idx, rows in peak_all.iterrows():
            peak_all.loc[idx, 'fdr'] = rows['P_Value'] * peak_num / rows['rank']

        # calculate the BHFDR
        peak_all = peak_all.sort_values('rank')
        peak_rank = peak_all['fdr'].to_list()
        for now_peak_num in range(int(peak_num)):
            now_fdr = peak_rank[now_peak_num]
            if now_peak_num + 1 < int(peak_num):
                now_min_fdr = min(peak_rank[now_peak_num + 1::])
                if now_fdr > now_min_fdr:
                    peak_rank[now_peak_num] = now_min_fdr
                else:
                    pass
            else:
                pass

        peak_all.loc[:, 'BHFDR'] = peak_rank
        del peak_all['fdr']
        peak_all = peak_all.sort_index()

        peak_all.to_csv(self.peak_txt_file, sep='\t', header=True, index=False)

        # save the results in bed format, start with 0
        # bed12 format: chrom / chromStart / chromEn / name / score / strand /
        # thickStart / thickEnd / itemRGB / blockCount / blockSize / blockStarts-
        peak_bed = peak_all.loc[:, ['transcripts', 'peak_start', 'peak_end', 'peak_num', 'mean_fold', 'max_site', 'P_Value']]
        peak_bed.loc[:, 'name'] = peak_bed['transcripts'] + '_peak_' + peak_bed['peak_num']
        peak_bed.loc[:, 'strand'] = '+'

        # peak_bed = peak_bed.apply(pd.to_numeric, errors='ignore')
        peak_bed= peak_bed.apply(self.safe_to_numeric)

        peak_bed.loc[:, 'peak_start'] = peak_bed.loc[:, 'peak_start'] - 1
        peak_bed = peak_bed.loc[peak_bed.loc[:, 'peak_start'] >= 0, :]
        peak_bed.loc[:, 'start'] = peak_bed.loc[:, 'peak_start'] * 3
        peak_bed.loc[:, 'end'] = peak_bed.loc[:, 'peak_end'] * 3 + 2

        bed_merge = peak_bed.loc[:, ['transcripts', 'start', 'end', 'name', 'mean_fold', 'strand']]
        bed_merge.columns = ['#transcripts', 'start', 'end', 'name', 'mean_fold', 'strand']
        bed_merge.to_csv(self.peak_bed_file, sep='\t', header=True, index=False)
        
        peak_sig_bed = peak_bed.loc[peak_bed.loc[:, 'P_Value'] < 0.05, :]
        bed_sig_merge = peak_sig_bed.loc[:, ['transcripts', 'start', 'end', 'name', 'mean_fold', 'strand']]
        bed_sig_merge.columns = ['#transcripts', 'start', 'end', 'name', 'mean_fold', 'strand']
        bed_sig_merge.to_csv(self.peak_sig_bed_file, sep='\t', header=True, index=False)

        # save the results in txt format contains the peak sequences
        # txt format: transcripts / gene_name / max_size / upstream sequence / peak sequence / downstream sequence

        peak_gene_name = peak_all.loc[:, 'transcripts'].unique().tolist()
        self.rpf_seq = self.rpf_seq.loc[self.rpf_seq.loc[:, 'name'].isin(peak_gene_name), :]
        rpf_seq_group = self.rpf_seq.groupby('name')

        peak_seq_list = []

        for idx, rows in peak_bed.iterrows():
            transcripts = rows['transcripts']
            gene_name = self.gene_dict[transcripts]
            peak_name = rows['name']

            up_start = rows['peak_start'] - self.upstream
            peak_start = rows['peak_start']
            peak_end = rows['peak_end']
            down_end = rows['peak_end'] + self.downstream

            gene_rpf = rpf_seq_group.get_group(transcripts)

            upstream_seq = gene_rpf.loc[(gene_rpf.from_tis >= up_start) & (gene_rpf.from_tis < peak_start), 'codon'].tolist()
            peak_seq = gene_rpf.loc[(gene_rpf.from_tis >= peak_start) & (gene_rpf.from_tis <= peak_end), 'codon'].tolist()
            downstream_seq = gene_rpf.loc[(gene_rpf.from_tis > peak_end) & (gene_rpf.from_tis <= down_end), 'codon'].tolist()
            max_site = rows['max_site']

            upstream_aa = Seq(''.join(upstream_seq)).translate()
            peak_aa = Seq(''.join(peak_seq)).translate()
            downstream_aa = Seq(''.join(downstream_seq)).translate()

            temp_peak_seq = [transcripts, gene_name, peak_name,
                             peak_start, peak_end, max_site,
                             ''.join(upstream_seq), ''.join(peak_seq), ''.join(downstream_seq),
                             str(upstream_aa), str(peak_aa), str(downstream_aa)]

            peak_seq_list.append(temp_peak_seq)
        
        self.peak_seq = pd.DataFrame(peak_seq_list, 
                                     columns=['transcripts', 'gene_name', 'peak_name',
                                              'peak_start', 'peak_end', 'max_site', 
                                              'upstream_nt', 'peak_nt', 'downstream_nt',
                                              'upstream_aa', 'peak_aa', 'downstream_aa'])
        
        self.peak_seq.to_csv(self.peak_seq_file, sep='\t', header=True, index=False)

        # # save the rpm of each peak
        # if self.rpm:
        #     self.output_peak_rpm()
        # else:
        #     pass

        # output the significant peaks enrichment ratio to self.peak_ratio_file
        peak_ratio_df = pd.concat(self.enrich_out, axis=0)
        peak_ratio_df.to_csv(self.peak_ratio_file, sep='\t', header=True, index=False)

        # save all the enrichment ratio of each gene
        if self.ratio_out:
            self.output_ratio()
        else:
            pass

    # define the function to output the matlab scripts
    def output_matlab(self):
        split_line = pd.DataFrame(["##############################################"])
        for mrna, ratio in self.all_ratio.items():
            ratio.index.name = mrna
            split_line.to_csv(self.peak_ratio_file, mode='a+', index=False, header=False)
            ratio.to_csv(self.peak_ratio_file, mode='a+', index=True, header=True)



"""
Author: 'rensc' 'rensc0718@163.com'
LastEditors: 'rensc' 'rensc0718@163.com'
LastEditTime: 2024-11-16 12:00:00
FilePath: \RiboParser\setup.py
Description:
"""

from setuptools import setup, find_packages

setup(
    name="RiboParser",
    version="0.1.6.1",
    packages=find_packages(),
    
    author="Ren Shuchao",
    author_email="rensc0718@163.com",
    description="A pipeline for ribosome profiling data analysis",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/renscq/RiboParser",
    license="GPL",
    platforms=['linux-x86_64'],
    classifiers=[
        "Programming Language :: Python :: 3.12",
        'Operating System :: POSIX :: Linux',
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.12",

    install_requires=[
        "numpy~=1.26.4",
        "pandas~=2.2.2",
        "pyarrow~=16.1.0",
        "polars~=0.20.31",
        "biopython~=1.78",
        "scipy~=1.12.0",
        "scikit-learn~=1.4.2",
        "statsmodels~=0.14.2",
        "pysam~=0.22.1",
        "joblib~=1.4.2",
        "interval~=1.0.0",
        "matplotlib~=3.8.4",
        "matplotlib-venn~=1.1.1",
        "seaborn~=0.13.2",
        "plotly~=5.22.0",
        "seqlogo~=5.29.9",
        "kaleido~=0.2.1",
    ],
    entry_points={
        "console_scripts": [
            # Ribo-quality
            "rpf_Reference=utils.rpf_Reference:main",
            "rpf_Check=utils.rpf_Check:main",
            "rpf_Digest=utils.rpf_Digest:main",
            "rpf_Offset=utils.rpf_Offset:main",
            "rpf_Offset_RSBM=utils.rpf_Offset_RSBM:main",
            "rpf_Density=utils.rpf_Density:main",
            "rpf_Merge=utils.rpf_Merge:main",
            "rpf_Periodicity=utils.rpf_Periodicity:main",
            "rpf_Metaplot=utils.rpf_Metaplot:main",
            "rpf_Coverage=utils.rpf_Coverage:main",
            "rpf_Corr=utils.rpf_Corr:main",
            "rpf_Quant=utils.rpf_Quant:main",
            # Ribo-pausing
            "rpf_Pausing=utils.rpf_Pausing:main",
            "rpf_Occupancy=utils.rpf_Occupancy:main",
            "rpf_CoV=utils.rpf_CoV:main",
            "rpf_Cumulative_CoV=utils.rpf_Cumulative_CoV:main",
            "rpf_CDT=utils.rpf_CDT:main",
            "rpf_CST=utils.rpf_CST:main",
            "rpf_Odd_Ratio=utils.rpf_Odd_Ratio:main",
            # Ribo-utils
            "rpf_Shuffle=utils.rpf_Shuffle:main",
            "rpf_Retrieve=utils.rpf_Retrieve:main",
            "rpf_Bam2bw=utils.rpf_Bam2bw:main",
            "rpf_Geneplot=utils.rpf_Geneplot:main",
            # RNA
            "rna_Density=utils.rna_Density:main",
            "rna_Offset=utils.rna_Offset:main",
            # SeRP
            "serp_overlap=utils.serp_overlap:main",
            "serp_peak=utils.serp_peak:main",
            "serp_properties=utils.serp_properties:main",
            # bedgraph
            "bg2meta=scripts.bedgraph.bg2meta:main",
            "rpm_smooth=scripts.bedgraph.rpm_smooth:main",
            # bowtie
            "merge_bwt_log=scripts.bowtie.merge_bwt_log:main",
            # fasta
            "fa_gc_sum=scripts.fasta.fa_gc_sum:main",
            "fa_len_flt=scripts.fasta.fa_len_flt:main",
            "fa_len_sum=scripts.fasta.fa_len_sum:main",
            "fa_split=scripts.fasta.fa_split:main",
            "line_feed=scripts.fasta.line_feed:main",
            "nt2aa=scripts.fasta.nt2aa:main",
            "rand_seq=scripts.fasta.rand_seq:main",
            "retrieve_seq=scripts.fasta.retrieve_seq:main",
            "revs=scripts.fasta.revs:main",
            # fastq
            "fq_len_flt=scripts.fastq.fq_len_flt:main",
            "fq_len_sum=scripts.fastq.fq_len_sum:main",
            "fq_split=scripts.fastq.fq_split:main",
            "fq_trim=scripts.fastq.fq_trim:main",
            "fq2fa=scripts.fastq.fq2fa:main",
            "fq2txt=scripts.fastq.fq2txt:main",
            "phred_quality=scripts.fastq.phred_quality:main",
            # merge_ribo
            "merge_cdt=scripts.merge_ribo.merge_cdt:main",
            "merge_coverage=scripts.merge_ribo.merge_coverage:main",
            "merge_cst=scripts.merge_ribo.merge_cst:main",
            "merge_digestion=scripts.merge_ribo.merge_digestion:main",
            "merge_dst_list=scripts.merge_ribo.merge_dst_list:main",
            "merge_length=scripts.merge_ribo.merge_length:main",
            "merge_metagene=scripts.merge_ribo.merge_metagene:main",
            "merge_occupancy=scripts.merge_ribo.merge_occupancy:main",
            "merge_odd_ratio=scripts.merge_ribo.merge_odd_ratio:main",
            "merge_offset_detail=scripts.merge_ribo.merge_offset_detail:main",
            "merge_offset=scripts.merge_ribo.merge_offset:main",
            "merge_pausing=scripts.merge_ribo.merge_pausing:main",
            "merge_period=scripts.merge_ribo.merge_period:main",
            "merge_quant=scripts.merge_ribo.merge_quant:main",
            "merge_saturation=scripts.merge_ribo.merge_saturation:main",
            # oligo
            "get_overlap_seq=scripts.oligo.get_overlap_seq:main",
            "get_tissue_freq=scripts.oligo.get_tissue_freq:main",
            "get_win_seq=scripts.oligo.get_win_seq:main",
            # ribocode
            "ribocode_bed_format=scripts.ribocode.ribocode_bed_format:main",
            # ribotish
            "ribotish_format=scripts.ribotish.ribotish_format:main",
            # rsem
            "merge_rsem=scripts.rsem.merge_rsem:main",
            # unix
            "dos2unix=scripts.unix.dos2unix:main",

        ],
    },

)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Project : riboParser
# @Script  : Shuffle.py


import pandas as pd
import polars as pl
import numpy as np
from . import RPFs


class Shuffle(object):

    def __init__(self, args):
        self.rpf_file = args.rpf
        
        self.gene = args.list

        self.shuffle_rpf = None
        self.seed = args.seed
        self.individual = args.individual

        self.output = args.output + '_rpf_shuffle.txt'

    def import_rpf(self):
        '''
        @Message  : import the rpf table.
        @Input    : self.rpf --> rpf table
                    self.site --> [E, P, A] site in ribosome
                    self.frame --> [0, 1, 2] frame in [E, P, A] site
                    self.gene --> gene list
                    self.rpf_num --> minimum rpf value
                    norm --> normalize the rpf value to RPM
        @Return   : self.raw_rpf --> rpf table contain all rpf value
                    self.sample_name --> samples name
                    self.sample_num --> groups sample number
                    self.merged_rpf --> merged rpf table
                    self.total_rpf_num --> total rpf number of each sample
        '''
        
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf_file,
                                      sample_num=None,
                                      sample_name=None,
                                      sites='P',
                                      frame='all',
                                      gene=self.gene,
                                      rpf_num=0)

        self.raw_rpf = rpf_results[0].to_pandas()
        self.sample_name = rpf_results[1]
        # self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        # self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        # self.high_gene = rpf_results[6]

        del rpf_results


    def shuffle_rpfs(self):
        '''
        @Message  : shuffle the rpf table.
        @Input    : self.raw_rpf --> rpf table contain all rpf value
                    self.seed --> random seed
        @Return   : self.shuffle_rpf --> shuffled rpf table
        '''

        sample_name = []
        for now_sp in self.sample_name:
            sample_name.extend([now_sp + '_f0', now_sp + '_f1', now_sp + '_f2'])

        raw_rpf_group = self.raw_rpf.groupby('name')
        
        if self.individual:
            raw_rpf_group_list = []

            now_num = 0
            for idx, now_group in raw_rpf_group:
                now_num += 1
                if now_num % 100 == 0:
                    print(f'Processing {now_num} genes.', flush=True)

                rpf_melt = pd.melt(now_group, id_vars=['name', 'now_nt', 'from_tis', 'from_tts', 'region', 'codon'], 
                                   value_vars=sample_name, 
                                   var_name='Frame', value_name='Count')
                rpf_melt['Sample'] = rpf_melt['Frame'].str[:-3]

                np.random.seed(self.seed)
                rpf_melt['Count'] = rpf_melt.groupby('Sample')['Count'].transform(lambda x: np.random.permutation(x))
                rpf_melt.drop('Sample', axis=1, inplace=True)

                rpf_melt = rpf_melt.pivot_table(index=['name', 'now_nt', 'from_tis', 'from_tts', 'region', 'codon'], 
                                     columns='Frame', values='Count', aggfunc='first').reset_index()

                raw_rpf_group_list.append(rpf_melt)

            print(f'Processing {now_num} genes.', flush=True)

        else:
            raw_rpf_group_list = []

            now_num = 0
            for idx, now_group in raw_rpf_group:
                now_num += 1
                if now_num % 100 == 0:
                    print(f'Processing {now_num} genes.', flush=True)

                rpf_melt = pd.melt(now_group, id_vars=['name', 'now_nt', 'from_tis', 'from_tts', 'region', 'codon'], 
                                   value_vars=sample_name, 
                                   var_name='Frame', value_name='Count')
                rpf_melt['Sample'] = rpf_melt['Frame'].str[:-3]
                rpf_melt['site'] = rpf_melt['now_nt'] + rpf_melt['Frame'].str[-1].astype(int)

                rpf_melt_count = rpf_melt.pivot_table(index='site', columns='Sample', values='Count', aggfunc='first').reset_index()

                # shuffle the rpf value
                np.random.seed(self.seed)
                rpf_melt_count_index = np.random.permutation(rpf_melt_count.index)
                rpf_melt_count = rpf_melt_count.reindex(rpf_melt_count_index).reset_index(drop=True)
                rpf_melt_count.drop('site', axis=1, inplace=True)
                rpf_melt_count = rpf_melt_count.melt(var_name='Sample', value_name='Count')
                
                rpf_melt['Count'] = rpf_melt_count['Count']
                rpf_melt.drop(['Sample', 'site'], axis=1, inplace=True)

                rpf_melt = rpf_melt.pivot_table(index=['name', 'now_nt', 'from_tis', 'from_tts', 'region', 'codon'], 
                                     columns='Frame', values='Count', aggfunc='first').reset_index()

                raw_rpf_group_list.append(rpf_melt)

            print(f'Processing {now_num} genes.', flush=True)

        self.shuffle_rpf = pd.concat(raw_rpf_group_list, axis=0)
        self.shuffle_rpf.reset_index(drop=True, inplace=True)


    def output_rpfs(self):
        '''
        @Message  : output the shuffled rpf table.
        @Input    : self.shuffle_rpf --> shuffled rpf table
                    self.output --> output file name
        @Return   : None
        '''
        
        self.shuffle_rpf = pl.DataFrame(self.shuffle_rpf)
        self.shuffle_rpf.write_csv(self.output, separator='\t')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2020/4/1 21:06
# @Project : mytools
# @Script  : fq_len_sum.py
# @Version : python 3.8.5
# @product : PyCharm
# @Author  : Rensc
# @E-mail  : rensc0718@163.com


import argparse
import gc
import time
from argparse import RawTextHelpFormatter


# get arguments and define the scripts usage
def parse_args():
    parser = argparse.ArgumentParser(
        description="This script is used to sum the reads length.",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s 1.0"
    )
    parser.add_argument(
        "-i", metavar="input", required=True, type=str, help="input file in fq format"
    )
    parser.add_argument(
        "-o", metavar="output", required=True, type=str, help="output file in fq format"
    )
    args = parser.parse_args()
    return args


# read the fastq file
def read_fq(in_fq_file, length_out_file):
    length_dict = {}
    sum_num = 0
    with open(in_fq_file, 'r') as fq_in:
        with open(length_out_file, 'w') as length_out:
            for line in fq_in:
                line2 = fq_in.__next__().strip()
                reads_length = len(line2)
                if length_dict.get(reads_length):
                    length_dict[reads_length] += 1
                else:
                    length_dict[reads_length] = 1
                fq_in.__next__()
                fq_in.__next__()

            length_out.writelines('length\treads_number\n')

            for length in sorted(length_dict.keys()):
                length_out.writelines(''.join([str(length), '\t', str(length_dict[length]), '\n']))
                sum_num += length_dict[length]
            length_out.writelines('Total reads number\t' + str(sum_num) + '\n')


# gc collect
def clear():
    print("Clear Memory.")
    gc.collect()
    print("All done.")


# get time
def now_time():
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))


# main
def main():
    now_time()
    args = parse_args()
    read_fq(args.i, args.o)
    clear()
    now_time()


# run
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_Cumulative_CoV.py


from .ribo import ArgsParser
from .ribo.Cumulative_CoV import *


def main():
    ArgsParser.now_time()
    print('Retrieve the RPFs with gene list.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.cumulative_cov_args_parser()
    rpfs = CumulativeCoV(args)

    print('Step2: Retrieve the gene RPFs.\n', flush=True)
    rpfs.retrieve_rpf()
    rpfs.rpf_to_rpm()

    print('Step3: Format the RPFs table.\n', flush=True)
    rpfs.melt_rpf_table()

    print('Step4: Calculate the cumulative CoV.\n', flush=True)
    rpfs.calc_cov()

    print('Step5: Output the cumulative CoV meta table.\n', flush=True)
    rpfs.merge_cov_table()

    print('Step5: Output the cumulative CoV table.\n', flush=True)
    rpfs.output_rpf_table()

    print('All done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : Periodicity.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2024/01/22 14:13:22
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


from . import RPFs
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
import polars as pl
import pandas as pd
import seaborn as sns
import math

class Periodicity(object):

    def __init__(self, args):
        # input and output
        self.transcript = args.transcript
        self.rpf = args.rpf
        self.output = args.output

        # filter the range around tis/tts
        self.rpf_num = args.min
        self.tis = args.tis
        self.tts = args.tts

        # for the high expression gene filter
        self.sample_num = 0
        self.raw_rpf = None
        self.high_gene = None

        self.period = None

    def import_rpf(self):
        '''
        @Message  : function for rpf import.
        @Input    : self.ribo --> rpf density file
                    self.tis --> tis range
                    self.tts --> tts range
                    self.rpf_num --> rpf number filter
        @Return   : output --> rpf density dataframe
        '''

        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites='P',
                                      frame='all',
                                      sample_num=None,
                                      sample_name=None,
                                      tis = self.tis,
                                      tts = self.tts,
                                      gene=self.transcript,
                                      rpf_num=self.rpf_num)
        
        self.raw_rpf = rpf_results[0].to_pandas()
        # self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        # self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7]

        del rpf_results

    def calc_3nt_period(self):
        '''
        @Message  : function for .
        @Input    : param --> description
        @Return   : output --> description
        @Flow     : step1 --> run
        '''

        raw_rpf = self.raw_rpf.drop(columns=["name", "now_nt", "from_tis", "from_tts", "region", "codon"])
        
        period = raw_rpf.apply(sum).to_frame()
        period = period.reset_index()
        period.columns = ["Name", "Count"]
        period["Sample"] = period['Name'].str[:-3]
        period["Frame"] = period['Name'].str[-1::]
        period["Ratio"] = period["Count"] / period.groupby("Sample")["Count"].transform("sum") * 100

        self.period = period[["Sample", "Frame", "Count", "Ratio"]]

    def output_meta(self):
        out_txt = self.output + "_periodicity.txt"
        self.period.to_csv(out_txt, sep='\t', index=False)

    def get_sub_plot_num(self):
        factor1 = int(self.sample_num ** 0.5)
        factor2 = int(self.sample_num ** 0.5)
        square = factor1 * factor2

        while not square >= self.sample_num:
            if factor2 > factor1:
                factor1 += 1
            else:
                factor2 += 1
            square = factor1 * factor2

        return factor1, factor2
    
    def draw_3nt_period_count(self):

        out_pdf = self.output + "_count_periodicity_plot.pdf"
        out_png = self.output + "_count_periodicity_plot.png"

        nrow, ncol = self.get_sub_plot_num()
        
        matplotlib.use('AGG')
        fig, axes = plt.subplots(nrows=nrow, ncols=ncol, figsize=(nrow * 4, ncol * 3), sharey=True)

        for (sample, group), ax in zip(self.period.groupby('Sample'), axes.flatten()):
            bars = group.plot(x='Frame', y='Count', kind='bar', ax=ax, title=sample, legend=False)
            ax.set_ylabel('Count')
            ax.set_xlabel('Frame')

            for bar in bars.patches:
                height = bar.get_height()
                ax.annotate(f'{height:.0f}', 
                            xy=(bar.get_x() + bar.get_width() / 2, height),
                            xytext=(0, 3),  # 3 points vertical offset
                            textcoords="offset points",
                            ha='center', va='bottom')
        
        plt.tight_layout()

        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png, dpi=300)

    def draw_3nt_period_ratio(self):

        out_pdf = self.output + "_ratio_periodicity_plot.pdf"
        out_png = self.output + "_ratio_periodicity_plot.png"

        nrow, ncol = self.get_sub_plot_num()

        matplotlib.use('AGG')
        fig, axes = plt.subplots(nrows=nrow, ncols=ncol, figsize=(nrow * 4, ncol * 3), sharey=True)

        for (sample, group), ax in zip(self.period.groupby('Sample'), axes.flatten()):
            bars = group.plot(x='Frame', y='Ratio', kind='bar', ax=ax, title=sample, legend=False)
            ax.set_ylabel('Ratio (%)')
            ax.set_xlabel('Frame')

            for bar in bars.patches:
                height = bar.get_height()
                ax.annotate(f'{height:.0f}', 
                            xy=(bar.get_x() + bar.get_width() / 2, height),
                            xytext=(0, 3),  # 3 points vertical offset
                            textcoords="offset points",
                            ha='center', va='bottom')
        
        plt.tight_layout()

        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png, dpi=300)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : ribo_parser.py


from ribo import ArgsParser
from ribo.Ribo import *


def format_seq(ribo_attr):
    rpf_list = []

    for name, isoform in ribo_attr.mrna_dict.items():
        rpf_dict = OrderedDict()
        # print(name + '\n')

        # trim the cds frame to fit the 3nt periodicity
        if isoform.cds_length % 3 != 0:
            shift_cds = isoform.cds_length % 3
            isoform.cds_length -= shift_cds
            isoform.utr3_length += shift_cds

        shift_5 = isoform.utr5_length % 3
        shift_3 = isoform.utr3_length % 3

        utr5_length = isoform.utr5_length - shift_5
        utr3_length = isoform.utr3_length - shift_3
        trim_length = isoform.length - shift_5 - shift_3

        rpf_dict['name'] = [name] * int(trim_length / 3)
        rpf_dict['now_nt'] = list(range(1 + shift_5, isoform.length - shift_3, 3))
        rpf_dict['from_tis'] = list(range(-utr5_length // 3, (trim_length - utr5_length) // 3))
        rpf_dict['from_tts'] = list(range((utr3_length - trim_length) // 3 + 1, utr3_length // 3 + 1))
        rpf_dict['region'] = ['5utr'] * (utr5_length // 3) + ['cds'] * (isoform.cds_length // 3) + ['3utr'] * (utr3_length // 3)

        if shift_3 != 0:
            trim_seq = isoform.seq[shift_5: -shift_3]
            rpf_dict['aa_list'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

            trim_rpf = isoform.rpf[shift_5: -shift_3]
            rpf_dict[ribo_attr.output + '_f0'] = trim_rpf[0::3]
            rpf_dict[ribo_attr.output + '_f1'] = trim_rpf[1::3]
            rpf_dict[ribo_attr.output + '_f2'] = trim_rpf[2::3]
        elif shift_3 == 0:
            trim_seq = isoform.seq[shift_5::]
            rpf_dict['aa_list'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

            trim_rpf = isoform.rpf[shift_5::]
            rpf_dict[ribo_attr.output + '_f0'] = trim_rpf[0::3]
            rpf_dict[ribo_attr.output + '_f1'] = trim_rpf[1::3]
            rpf_dict[ribo_attr.output + '_f2'] = trim_rpf[2::3]
        try:
            rpf = pd.DataFrame(rpf_dict)
            rpf_list.append(rpf)
        except ValueError:
            print('Checking the input Arguments.\n'.format(gene_name=name), flush=True)

    total_rpf_df = pd.concat(rpf_list, ignore_index=True)
    # total_rpf_df["sum"] = total_rpf_df[['frame_1', 'frame_2', 'frame_3']].sum(axis=1)
    total_rpf_df.to_csv(ribo_attr.output + "_rpf.txt", sep='\t', index=False)


def main():
    ArgsParser.now_time()
    print('\nParse the RIBO-SEQ data.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.ribo_args_parser()
    ribo_attr = Ribo(args)

    print('\nStep2: Import the p-site offset.\n', flush=True)
    ribo_attr.read_offset()

    print('\nStep3: Import the transcripts annotation.\n', flush=True)
    ribo_attr.read_transcript()

    print('\nStep4: Import the BAM file.\n', flush=True)
    ribo_attr.read_bam()

    print('\nStep5: Format the in-frame RPFs.\n', flush=True)
    format_seq(ribo_attr)

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_metaplot.py


from .ribo import ArgsParser
from .ribo import Metaplot


def main():
    ArgsParser.now_time()
    print('\nDraw the metaplot.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.metaplot_args_parser()
    meta = Metaplot.Metaplot(args)

    print('\nStep2: Import the RPFs file.\n', flush=True)
    meta.gene_anno()
    meta.read_rpf()
    meta.output_merge_meta()

    print('\nStep3: Draw the metaplot.\n', flush=True)
    if args.mode == "bar":
        meta.draw_bar_metaplot()
    elif args.mode == "line":
        meta.draw_line_metaplot()

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_occupancy.py


import pandas as pd
import argparse


def merge_occupancy_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the occupancy files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for occupancy files (e.g., '*_codon_occupancy.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name (default: prefix + _codon_occupancy.txt).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_occupancy_files(occupancy_list):
    '''
    @Message  : retrieve the occupancy table.
    @Input    : list --> pattern for occupancy files (e.g., '*_codon_occupancy.txt')
    @Return   : 
                output --> output dataframe contain the occupancy
    @Flow     : step1 --> retrieve the occupancy
    '''

    # make the dataframe
    occupancy_df = pd.DataFrame()

    # for each occupancy file
    for occupancy_file in occupancy_list:
        # import the occupancy file
        occupancy = pd.read_csv(occupancy_file, sep='\t', index_col=False)
        occupancy.sort_values(by=['Abbr', 'Codon'], ascending=True, inplace=True, ignore_index=True)

        # merge the gene occupancy file
        if occupancy_df.empty:
            occupancy_df = occupancy
        else:
            # delete the columns ['AA', 'Abbr.']
            occupancy = occupancy.drop(['AA', 'Abbr'], axis=1)
            occupancy_df = pd.merge(left=occupancy_df, right=occupancy, how='outer', on='Codon')

    # reorder the occupancy columns with 
    # ['codon', 'AA', 'Abbr.', *_density, *_absolute_occupancy, *_relative_occupancy]
    occupancy_codon = occupancy_df[['Codon', 'AA', 'Abbr']]
    occupancy_density = occupancy_df.filter(regex='_density')
    occupancy_abs = occupancy_df.filter(regex='_absolute_occupancy')
    occupancy_rel = occupancy_df.filter(regex='_relative_occupancy')

    occupancy_merge = pd.concat([occupancy_codon, occupancy_density, occupancy_abs, occupancy_rel], axis=1)

    return occupancy_merge


def output_table(occupancy_merge, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the occupancy
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    occupancy_merge.to_csv(output_file + '_codon_occupancy.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_occupancy_args_parser()

    print('Step2: import the occupancy file.', flush=True)
    occupancy_df = import_occupancy_files(args.list)

    print('Step3: output the occupancy table.', flush=True)
    output_table(occupancy_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : CST.py


import sys
from collections import OrderedDict
import pandas as pd
import polars as pl
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
# import xlsxwriter
from . import RPFs


class CodonTable(object):
    def __init__(self):
        """
        commonly contains the follow elements:

        codon number, frequency, properties
        ribo-seq number, rna-seq number
        valid codon number
        codon density proportion
        """
        self.codon = None
        self.aa = None
        self.abbr = None
        self.number = 0
        self.frequency = 0
        self.cai = 0
        self.cbi = 0
        self.rscu = 0
        self.properties = None

        self.rpf_codon_sum = 0
        self.rpf_codon_val = 0
        self.rpf_sum = 0
        self.rpf_proportion = 0

        self.rna_codon_sum = 0
        self.rna_codon_val = 0
        self.rna_sum = 0
        self.rna_proportion = []

        self.cst = []


class GeneTable(object):
    def __init__(self, name):
        """
        gene: gene name
        codon count dict: (codon: number)
        codon_num: total codon number of each gene
        rpf, rna reads: Ribo-seq and RNA-seq reads density
        density: (TE = rpf/rna)
        elongation rate: [codon_num / Σ(count * CST))]
        initiate rate: (elongation rate * density)
        """
        self.gene = name
        self.length = 0
        self.codon_num = 0
        self.elongate_rate = 0
        self.rpf = 0
        self.rna = 0
        self.density = 0
        self.initiate_rate = 0
        self.codon_count = OrderedDict()


class CodonSelectiveTime(object):

    def __init__(self, args):
        """
        receive arguments and set the self parameters
        """
        self.gene = args.list
        self.rpf_file = args.rpf
        self.rna_file = args.rna

        self.rpf_sample = None
        self.rna_sample = None

        # for the high expression gene filter
        self.gene_list = args.list
        self.rpf_num = args.min
        
        self.rpf_gene_sum = None
        self.rna_gene_sum = None

        self.merge_rpf = None
        self.merge_rna = None

        self.high_rpf = None
        self.high_rna = None

        self.rpf_gene = None
        self.rna_gene = None

        self.overlap_gene = None
        self.high_gene = None

        self.high_rpf_num = 0
        self.high_rna_num = 0

        self.total_rpf_num = 0
        self.total_rna_num = 0

        # filter the stable codon coverage range
        self.site = args.site
        self.frame = args.frame
        self.scale = args.scale

        self.tis = args.tis
        self.tts = args.tts
        self.times = args.times

        # table for codon selection time
        self.stop = args.stop
        self.codon_dict = RPFs.codon_table()[0]
        
        if self.stop:
            del self.codon_dict['TAA']
            del self.codon_dict['TAG']
            del self.codon_dict['TGA']

        self.codon_table = OrderedDict()

        self.gene_table = OrderedDict()
        self.gene_length_table = None
        self.gene_codon_table = None

        self.detail_cst = None
        self.merge_cst = None
        self.absolute_cst = None
        self.relative_cst = None
        self.output = args.output


    def import_density(self):
        '''
        @Message  : function for import the Ribo-seq and RNA-seq density file.
                    By default, the two data names are different.
        @Input    : rpf_file --> rpf/rna file
                    sample_num --> sample number
                    sample_name --> sample name
                    sites --> filter the stable codon coverage range
                    frame --> filter the stable codon coverage frame
                    gene --> high expression gene list
                    rpf_num --> filter the low expression gene
                    tis --> filter the TIS region start
                    tts --> filter the TTS region stop
        @Return   : rpf_sample --> samples name
                    high_rpf --> high expression gene density
                    rpf_gene --> high expression gene name
                    total_rpf_num --> total rpf/reads
        @Flow     : step1 --> import the rpf density
                    step2 --> import the rna density
                    step3 --> check the samples number
        '''
        
        # import the rpf density
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf_file, sample_num=None,
                                      sample_name=None, sites=self.site, frame=self.frame,
                                      gene=None, rpf_num=1, tis=self.tis, tts=self.tts)
        self.rpf_sample = rpf_results[1]
        self.merge_rpf = pl.DataFrame(rpf_results[3])
        self.total_rpf_num = rpf_results[4]
        self.rpf_gene_sum = rpf_results[5]
        self.rpf_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7].copy()

        del rpf_results

        print('', flush=True)

        # import the rna density
        rna_results = RPFs.import_rpf(rpf_file=self.rna_file, sample_num=None,
                                      sample_name=None, sites=self.site, frame=self.frame,
                                      gene=None, rpf_num=1, tis=self.tis, tts=self.tts)
        self.rna_sample = rna_results[1]
        self.merge_rna = pl.DataFrame(rna_results[3])
        self.total_rna_num = rna_results[4]
        self.rna_gene_sum = rna_results[5]
        self.rna_gene = rna_results[6]
        # self.high_rna = rna_results[7].copy()

        del rna_results
        
        # check the samples number
        rpf_sample_num = len(self.rpf_sample)
        rna_sample_num = len(self.rna_sample)

        if rpf_sample_num != rna_sample_num:
            print('''
                    The number of samples in Ribo-seq and RNA-seq is not equal, please check the input density file.
                  ''', flush=True)
            sys.exit(1)


    def merge_rpf_rna(self):
        '''
        @Message  : get the overlap gene from rpf and mrna density file.
                    only these genes will submit to follow analysis.
                    PS: the gene name should be the same in the two density file.
        @Input    : rpf_gene/rna_gene --> high expression gene name
                    high_rpf/high_rna --> high expression gene density
        @Return   : high_rpf/high_rna --> filtered high expression gene density
                    high_rpf_num/high_rna_num --> total rpf/reads
                    high_gene --> merged high expression gene density
        @Flow     : step1 --> run
        '''
        
        # get the overlap gene from rpf and mrna density file
        overlap_gene = set(self.rpf_gene) & set(self.rna_gene)

        self.high_rpf = self.high_rpf.loc[self.high_rpf['name'].isin(overlap_gene)]
        self.high_rna = self.high_rna.loc[self.high_rna['name'].isin(overlap_gene)]

        self.high_rpf_num = self.high_rpf[self.rpf_sample].sum()
        self.high_rna_num = self.high_rna[self.rna_sample].sum()

        # reset the index and sort the high expression gene
        self.high_rpf.reset_index(drop=True, inplace=True)
        self.high_rna.reset_index(drop=True, inplace=True)

        # self.high_rpf = self.high_rpf.sort_values(by=['name', 'now_nt'])
        # self.high_rna = self.high_rna.sort_values(by=['name', 'now_nt'])
        # self.high_gene = pd.merge(self.high_rpf, self.high_rna, on=['name', 'now_nt', 'from_tis', 'from_tts', 'region', 'codon'])
        # self.high_gene = pd.concat([self.high_rpf, self.high_rna.iloc[:, 6::]], axis=1)

        # self.high_rpf = pl.DataFrame(self.high_rpf)
        # self.high_rna = pl.DataFrame(self.high_rna)


    def make_codon_table(self, rpf_sample=None):

        '''
        @Message  :  make the codon message table contains: codon, name, amino-acid.
        @Input    : codon_dict --> codon message
                    rpf_sample --> sample name
        @Return   : codon_table --> codon message table
        @Flow     : step1 --> create the class for each codon
                    step2 --> merge the codon message
        '''
        
        sp_codon_table = OrderedDict()

        for codon, mess in self.codon_dict.items():
            sp_codon_table[codon] = CodonTable()
            sp_codon_table[codon].codon = codon
            sp_codon_table[codon].aa = mess[0]
            sp_codon_table[codon].abbr = mess[1]

        self.codon_table[rpf_sample] = sp_codon_table


    def filter_rpf_rna_expr(self, rpf_sample=None, rna_sample=None):
        '''
        @Message  : filter the high expression gene > self.rpf_num.
        @Input    : high_rpf --> high expression gene density
                    high_rna --> high expression gene density
                    rpf_sample --> sample name
                    rna_sample --> sample name
                    self.rpf_num --> filter the low expression gene
        @Return   : high_rpf --> filtered high expression gene density
                    high_rna --> filtered high expression gene density
                    overlap_gene --> high expression gene name in the Ribo-seq and RNA-seq data
                    high_rpf_num --> total rpf/reads
                    high_rna_num --> total rna/reads
        @Flow     : step1 --> filter the high expression gene with rpf > self.rpf_num
                    step2 --> check the high expression gene
                    step3 --> filter the high expression gene density
                    step4 --> summary the total rpf/reads and rna/reads
        '''
        
        # filter the high expression gene > self.rpf_num

        rpf_genes = self.rpf_gene_sum[self.rpf_gene_sum[rpf_sample] >= self.rpf_num].index
        rna_genes = self.rna_gene_sum[self.rna_gene_sum[rna_sample] >= self.rpf_num].index
        self.overlap_gene = set(rna_genes) & set(rpf_genes)

        # check the high expression gene
        if len(self.overlap_gene) == 0:
            print('No fitted gene in the Ribo-seq and RNA-seq data.', flush=True)
            sys.exit(1)
        else:
            print('The number of gene in the Ribo-seq and RNA-seq data is: {}'.format(len(self.overlap_gene)), flush=True)

        # filter the high expression gene density
        self.high_rpf = self.merge_rpf.filter(pl.col('name').is_in(self.overlap_gene))
        self.high_rna = self.merge_rna.filter(pl.col('name').is_in(self.overlap_gene))

        # self.high_rpf = self.high_rpf[self.high_rpf['name'].isin(self.overlap_gene)]
        # self.high_rna = self.high_rna[self.high_rna['name'].isin(self.overlap_gene)]

        # summary the total rpf/reads and rna/reads
        self.high_rpf_num = self.high_rpf[self.rpf_sample].sum()
        self.high_rna_num = self.high_rna[self.rna_sample].sum()

        # summary the gene length and codon number
        self.gene_length_table = self.high_rpf.to_pandas().groupby('name').size()
        self.gene_codon_table = self.high_rpf.to_pandas().groupby(['name', 'codon']).size()


    def make_gene_table(self, rpf_sample=None, rna_sample=None):

        '''
        @Message  : make high expression gene table with the following parameters:.
        @Input    : high_rpf --> high expression gene density
                    rpf_sample --> sample name
                    rna_sample --> sample name
        @Return   : gene_table --> gene table contain the [codon count, gene length, codon number, rpf, rna, density, elongate rate, initiate rate]
        @Flow     : step1 --> create the class for each gene
                    step2 --> summary the [codon count, gene length, codon number, rpf, rna, density]
                    step3 --> calculate the density with the formula: density = rpf / rna
        '''

        # make the gene table
        self.gene_table[rpf_sample] = OrderedDict()
        for gene in self.overlap_gene:
            self.gene_table[rpf_sample][gene] = GeneTable(gene)
            self.gene_table[rpf_sample][gene].codon_count = dict(self.gene_codon_table[gene])
            self.gene_table[rpf_sample][gene].length = self.gene_length_table[gene] * 3
            self.gene_table[rpf_sample][gene].codon_num = self.gene_length_table[gene]
            self.gene_table[rpf_sample][gene].rpf = float(self.rpf_gene_sum[rpf_sample].loc[gene])
            self.gene_table[rpf_sample][gene].rna = float(self.rna_gene_sum[rna_sample].loc[gene])
            self.gene_table[rpf_sample][gene].density = self.gene_table[rpf_sample][gene].rpf / self.gene_table[rpf_sample][gene].rna


    def ribo_codon_frequency(self, rpf_sample=None):

        '''
        @Message  : calculate the mean rpf density of codon.
        @Input    : high_rpf --> high expression gene density
                    codon_dict --> codon message
                    rpf_sample --> sample name
        @Return   : codon_table --> codon message table contain the [rpf codon sum, rpf codon val, rpf sum, rpf proportion]
        @Flow     : step1 --> create the class for each codon
                    step2 --> summary the [rpf codon sum, rpf codon val, rpf sum, rpf proportion]
                                rpf_codon_proportion = Σ(codon) / Σ(gene count)
                    step3 --> merge the codon message
        '''
        
        for codon in self.codon_dict.keys():

            codon_flt = self.high_rpf.filter(pl.col("codon") == codon)
            # codon_flt = self.high_rpf[self.high_rpf["codon"] == codon]
            codon_sum = codon_flt[rpf_sample].sum()
            codon_proportion = float(codon_sum / self.high_rpf_num[rpf_sample][0])

            self.codon_table[rpf_sample][codon].rpf_codon_sum = len(codon_flt)
            self.codon_table[rpf_sample][codon].rpf_codon_val = (codon_flt[rpf_sample] > 0).sum()
            self.codon_table[rpf_sample][codon].rpf_sum = codon_sum
            self.codon_table[rpf_sample][codon].rpf_proportion = codon_proportion


    def mrna_codon_frequency(self, rpf_sample=None, rna_sample=None, times=0, initiate_rate=None):

        '''
        @Message  : calculate the mean rna density of codon.
        @Input    : high_rna --> high expression gene density
                    codon_dict --> codon message
                    rpf_sample --> sample name
                    rna_sample --> sample name
                    times --> iterative times
                    initiate_rate --> normalize the rna density with initiate_rate
        @Return   : codon_table --> codon message table contain the [rna codon sum, rna codon val, rna sum, rna proportion, cst]
        @Flow     : step1 --> create the class for each codon
                    step2 --> summary the [rna codon sum, rna codon val, rna sum, rna proportion] with the formula:
                                mrna_codon_proportion = Σ(codon) / Σ(gene count)
                    step3 --> calculate the cst with the formula: 
                                cst = rpf_codon_proportion / mrna_codon_proportion
                    step4 --> merge the codon message
        '''

        if initiate_rate:
            high_rna = self.high_rna.with_columns(
                # self.high_rna[rna_sample] * self.high_rna['name'].apply(lambda x: self.gene_table[rpf_sample][x].initiate_rate if x in self.gene_table[rpf_sample].keys() else 1))
                # Calling `map_elements` without specifying `return_dtype` can lead to unpredictable results. Specify `return_dtype` to silence this warning
                self.high_rna[rna_sample] * self.high_rna['name'].apply(
                    lambda x: self.gene_table[rpf_sample][x].initiate_rate if x in self.gene_table[rpf_sample].keys() else 1,
                    return_dtype=pl.Float32))
            
        else:
            high_rna = self.high_rna

        for codon in self.codon_dict.keys():
            codon_flt = high_rna.filter(pl.col("codon") == codon)
            # codon_flt = high_rna[high_rna["codon"] == codon]
            codon_sum = codon_flt[rna_sample].sum()
            self.codon_table[rpf_sample][codon].rna_codon_sum = len(codon_flt)
            self.codon_table[rpf_sample][codon].rna_codon_val = (codon_flt[rna_sample] > 0).sum()
            self.codon_table[rpf_sample][codon].rna_sum = codon_sum

            rna_proportion = float(codon_sum / self.high_rna_num[rna_sample][0])
            self.codon_table[rpf_sample][codon].rna_proportion.append(rna_proportion)

            try:
                cst = self.codon_table[rpf_sample][codon].rpf_proportion / self.codon_table[rpf_sample][codon].rna_proportion[times]
            except ZeroDivisionError:
                cst = 0

            self.codon_table[rpf_sample][codon].cst.append(cst)


    def codon_elongate_rate(self, rpf_sample=None, codon_num=None, codon_dict=None, times=0):

        '''
        @Message  : calculate the elongate rate for each codon of gene.
        @Input    : rpf_sample --> sample name
                    codon_num --> total codon number of each gene
                    codon_dict --> codon count
                    times --> iterative times
        @Return   : elongate_rate --> elongate rate for each codon of gene
        @Flow     : step1 --> calculate the codon density
                    step2 --> calculate the elongate rate with the formula: elongate_rate = codon_num / Σ(count * CST))
        '''
        
        codon_density = 0
        for codon, number in codon_dict.items():
            try:
                codon_density += codon_dict[codon] * self.codon_table[rpf_sample][codon].cst[times]
            except KeyError:
                pass

        elongate_rate = codon_num / codon_density
        return elongate_rate
    

    def compute_cst_iterative(self, rpf_sample=None, rna_sample=None):

        '''
        @Message  : compute cst iteratively.
        @Input    : rpf_sample --> sample name
                    rna_sample --> sample name
                    self.times --> iterative times
                    self.overlap_gene --> high expression gene name in the Ribo-seq and RNA-seq data
        @Return   : gene_table --> gene table contain the [elongate rate, initiate rate]
        @Flow     : step1 --> calculate the elongate rate for each codon of gene
                                formula: elongate_rate = codon_num / Σ(count * CST))
                    step2 --> calculate the initiate rate for each gene
                                formula: initiate_rate = elongate_rate * density
        '''
        
        for iters in range(0, self.times, 1):
            print('Times of iteration: {}'.format(iters + 1), flush=True)
            for gene in self.overlap_gene:
                self.gene_table[rpf_sample][gene].elongate_rate = self.codon_elongate_rate(
                    rpf_sample=rpf_sample,
                    codon_num=self.gene_table[rpf_sample][gene].codon_num,
                    codon_dict=self.gene_table[rpf_sample][gene].codon_count,
                    times=iters)
                
                self.gene_table[rpf_sample][gene].initiate_rate = self.gene_table[rpf_sample][gene].elongate_rate * self.gene_table[rpf_sample][gene].density

            # self.high_rna = self.high_rna.with_columns(
            #     self.high_rna[self.rna_sample] * self.high_rna['name'].apply(
            #     lambda x: self.gene_table[x].initiate_rate if x in self.gene_table.keys() else 1))

            self.mrna_codon_frequency(rpf_sample=rpf_sample, rna_sample=rna_sample, times=iters + 1, initiate_rate=True)
    

    def calc_cst(self):
        
        '''
        @Message  : calculate the mean reads density of codon.
        @Input    : high_rpf --> high expression gene density
                    high_rna --> high expression gene density
                    self.rpf_sample --> sample name
                    self.rna_sample --> sample name
        @Return   : codon_cst_table --> codon cst table contain the 
                                        [rpf codon sum, rpf codon val, rpf sum, rpf proportion]
                                        [rna codon sum, rna codon val, rna sum, rna proportion, cst]
        @Flow     : step1 --> make the codon message table
                    step2 --> calculate the mean rpf density of codon
                    step3 --> calculate the mean rna density of codon
                    step4 --> make the gene table
                    step5 --> calculate the elongate rate for each codon of gene
                    step6 --> compute cst iteratively
        '''
        
        for rpf_sample, rna_sample in zip(self.rpf_sample, self.rna_sample):
            print('\nCalculate the codon selection time for samples: {rpf} + {rna}'.format(rpf=rpf_sample, rna=rna_sample), flush=True)

            self.make_codon_table(rpf_sample=rpf_sample)
            self.filter_rpf_rna_expr(rpf_sample=rpf_sample, rna_sample=rna_sample)
            self.make_gene_table(rpf_sample=rpf_sample, rna_sample=rna_sample)

            self.ribo_codon_frequency(rpf_sample=rpf_sample)
            self.mrna_codon_frequency(rpf_sample=rpf_sample, rna_sample=rna_sample)

            if self.times > 0:
                print('Calculate the codon selection time iterative.', flush=True)
                self.compute_cst_iterative(rpf_sample=rpf_sample, rna_sample=rna_sample)
            else:
                print('Skip the iterative calculation of codon selection time.', flush=True)


    @staticmethod
    def scale_method(scale, cst):
        if scale == 'minmax':
            # relative_cst = (cst - cst.min()) / (cst.max() - cst.min())
            relative_cst = cst / cst.max()
        elif scale == 'zscore':
            relative_cst = (cst - cst.mean()) / cst.std()
        else:
            print('Unknown scale method.', flush=True)
            sys.exit()
        return relative_cst
    

    def format_cst_results(self):
        '''
        @Message  : format the codon selection time results.
        @Input    : self.codon_table --> codon message table
                    self.rpf_sample --> sample name
                    self.codon_dict --> codon message
        @Return   : self.detail_cst --> details of codon selection time results
                    self.absolute_cst --> absolute codon selection time results
                    self.relative_cst --> relative codon selection time results
                    self.merge_cst --> merged codon selection time results
        @Flow     : step1 --> merge all codon selection time
                    step2 --> retrieve the absolute and relative codon selection time
                    step3 --> retrieve the absolute and relative codon selection time
                    step4 --> merge the absolute and relative codon selection time
        '''
        
        # set the column name
        column_name = ['Codon', 'AA', 'Abbr', 'rpf_codon_val', 'rpf_sum', 'rpf_proportion',
                       'rna_codon_val', 'rna_sum', 'rna_proportion_first', 'rna_proportion_last']

        cst_name = ['absolute_cst_' + str(iters) for iters in range(self.times + 1)]
        relative_cst_name = ['relative_cst_' + str(iters) for iters in range(self.times + 1)]

        # merge all codon selection time
        detail_cst = OrderedDict()

        for rpf_sample in self.rpf_sample:
            detail_cst[rpf_sample] = OrderedDict()

            for codon in self.codon_dict.keys():
                detail_cst[rpf_sample][codon] = [self.codon_table[rpf_sample][codon].aa,
                                                 self.codon_table[rpf_sample][codon].abbr,
                                                 self.codon_table[rpf_sample][codon].rpf_codon_val,
                                                 self.codon_table[rpf_sample][codon].rpf_sum,
                                                 self.codon_table[rpf_sample][codon].rpf_proportion,
                                                 self.codon_table[rpf_sample][codon].rpf_codon_val,
                                                 self.codon_table[rpf_sample][codon].rna_sum,
                                                 self.codon_table[rpf_sample][codon].rna_proportion[0],
                                                 self.codon_table[rpf_sample][codon].rna_proportion[-1]
                                                ]
                detail_cst[rpf_sample][codon].extend(self.codon_table[rpf_sample][codon].cst)

            detail_cst_df = pd.DataFrame.from_dict(detail_cst[rpf_sample]).transpose().reset_index(drop=False, inplace=False)
            detail_cst_df.columns = column_name + cst_name

            detail_rel_cst_df = self.scale_method(self.scale, detail_cst_df.loc[:, cst_name])
            detail_rel_cst_df.columns = relative_cst_name

            detail_cst_df = pd.concat([detail_cst_df, detail_rel_cst_df], axis=1)
            detail_cst_df['sample'] = rpf_sample

            if self.detail_cst is None:
                self.detail_cst = detail_cst_df
            else:
                self.detail_cst = pd.concat([self.detail_cst, detail_cst_df], axis=0)

        # retrieve the absolute and relative codon selection time
        self.absolute_cst = self.detail_cst.pivot_table(index=['Codon', 'AA', 'Abbr'],
                                                        columns='sample',
                                                        values=cst_name[-1],
                                                        aggfunc='first').reset_index(drop=False, inplace=False)

        # retrieve the absolute and relative codon selection time
        self.relative_cst = self.detail_cst.pivot_table(index=['Codon', 'AA', 'Abbr'],
                                                        columns='sample',
                                                        values=relative_cst_name[-1],
                                                        aggfunc='first').reset_index(drop=False, inplace=False)

        abs_column_name = ['Codon', 'AA', 'Abbr'] + [i + '_absolute_cst' for i in self.rpf_sample]
        relative_column_name = ['Codon', 'AA', 'Abbr'] + [i + '_relative_cst' for i in self.rpf_sample]
        
        self.absolute_cst.columns = abs_column_name
        self.relative_cst.columns = relative_column_name

        self.merge_cst = pd.merge(left=self.absolute_cst, right=self.relative_cst, how='outer', on=['Codon', 'AA', 'Abbr'])
        self.merge_cst.sort_values(by=['Abbr', 'Codon'], ascending=True, inplace=True, ignore_index=True)


    def output_cst(self):
        # cst_writer = pd.ExcelWriter(self.output + '_absolute_cst.xlsx', engine='xlsxwriter')
        # with pd.ExcelWriter(self.output + '_codon_selection_time.xlsx',) as cst_writer:
        #     self.absolute_cst.to_excel(cst_writer, sheet_name='absolute_cst', index=False)
        #     self.relative_cst.to_excel(cst_writer, sheet_name='relative_cst', index=False)

        self.detail_cst.to_csv(self.output + '_iterative_codon_selection_time.txt', sep='\t', index=False)
        self.merge_cst.to_csv(self.output + '_codon_selection_time.txt', sep='\t', index=False)


    def draw_cst_line(self):
        import plotly.express as px
        out_pdf = self.output + "_cst_line.pdf"
        out_png = self.output + "_cst_line.png"
        
        matplotlib.use('AGG')
        rel_cst = self.relative_cst.melt(id_vars=['Codon', 'AA'],
                                         var_name='iteration',
                                         value_name='cst',
                                         value_vars=['cst_' + str(iters) for iters in range(self.times + 1)])

        fig = px.line(rel_cst, x = "iteration", y = "cst", title = 'codon selection time')
        # fig.show()
        fig.write_image(out_pdf)
        fig.write_image(out_png)


    def draw_cst_corr(self):
        out_pdf = self.output + "_cst_corrplot.pdf"
        out_png = self.output + "_cst_corrplot.png"

        cst_corr = self.merge_cst.filter(regex='_absolute_cst').astype('float').corr()
        cst_corr.to_csv(self.output + '_cst_corr.txt', sep='\t', index=True)

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(6 + 0.2 * len(self.rpf_sample),
                                  8 + 0.2 * len(self.rpf_sample)),
                                  dpi=300)
        sns.heatmap(data=cst_corr, 
                    annot=True, 
                    fmt=".4f", 
                    cmap="vlag",
                    linewidths=.01)
        plt.tight_layout()
        # plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()


    def draw_cst_heat(self):
        # output the absolute cst
        out_pdf = self.output + "_cst_heatplot.pdf"
        out_png = self.output + "_cst_heatplot.png"

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(10 + 0.5 * len(self.rpf_sample), 20), dpi=300)
        sns.heatmap(data=self.merge_cst.filter(regex='_absolute_cst').astype('float'),
                    yticklabels=self.merge_cst['Codon'] + '[' + self.merge_cst['Abbr'].astype(str) + ']',
                    annot=True, 
                    fmt=".4f", 
                    cmap="vlag", 
                    linewidths=0)
        plt.tight_layout()
        # plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()
